import abc
import importlib
import typing

from keylime import config

from . import tests

# This module defines Policy for testing measured boot logs.
# This module also implements a registry of policies and a few trivial
# policies.

# RefState is a succinct description of what is expected to be found
# in a measured boot log.  A Policy maps one of these to the
# corresponding Test.
RefState = typing.Mapping[str, tests.Data]


# Policy maps RefState expressed in some convenient form into the
# corresponding Test to apply to the log.  The log is Python data
# corresponding to JSON, and is the result of parsing and enriching.
class Policy(metaclass=abc.ABCMeta):
    """Policy can compile RefState into a Test"""

    @abc.abstractmethod
    def get_relevant_pcrs(self) -> typing.FrozenSet[int]:
        """Reveal the set of relevant PCR indices"""
        raise NotImplementedError

    @abc.abstractmethod
    def refstate_to_test(self, refstate: RefState) -> tests.Test:
        """Convert the given RefState into a precise Test"""
        raise NotImplementedError

    def evaluate(self, refstate: RefState, eventlog: tests.Data) -> str:
        """Evaluate and return the reason for rejection or empty string for accept"""
        tester = self.refstate_to_test(refstate)
        return tester.why_not({}, eventlog)


class AcceptAll(Policy):
    """Policy that accepts all eventlogs"""

    def get_relevant_pcrs(self) -> typing.FrozenSet[int]:
        return set()

    def refstate_to_test(self, refstate: RefState) -> tests.Test:
        return tests.AcceptAll()


class RejectAll(Policy):
    """Policy that rejects all eventlogs"""

    def get_relevant_pcrs(self) -> typing.FrozenSet[int]:
        return set()

    def refstate_to_test(self, refstate: RefState) -> tests.Test:
        return tests.RejectAll("reject all")


def _mkreg() -> typing.Mapping[str, Policy]:
    return {}


_registry = _mkreg()


def register(name: str, policy: Policy):
    """Remember the given policy under the given name"""
    _registry[name] = policy


register("accept-all", AcceptAll())
register("reject-all", RejectAll())


def get_policy_names() -> typing.Tuple[str, ...]:
    """Return the list of policy names"""
    return list(_registry.keys())


def get_policy(name: str) -> Policy:
    """Returns the Policy with the given name, None if there is none"""
    return _registry.get(name)


def refstate_to_test(policy_name: str, refstate: RefState) -> tests.Test:
    """Compiles the given RefState into the Test prescribed by the named policy"""
    policy = get_policy(policy_name)
    if policy is None:
        raise Exception(f"there is no policy named {policy_name!a}")
    return policy.refstate_to_test(refstate)


def evaluate(policy_name: str, refstate: RefState, eventlog: tests.Data) -> str:
    """Evaluate the given eventlog using given refstate and policy

    Returns either:
    (a) an empty string to signal a good result or
    (b) a non-empty string identifying something wrong.
    """
    tester = refstate_to_test(policy_name, refstate)
    return tester.why_not({}, eventlog)


imports = config.getlist("verifier", "measured_boot_imports")
# print(f'importing {imports!r}, __package__={__package__!r}')
if imports:
    for imp in imports:
        if imp:
            importlib.import_module(imp, __package__)



"""
SPDX-License-Identifier: Apache-2.0
Copyright 2021 Red Hat, Inc.
"""

import unittest

from keylime import json as keylime_json


class JSON_Test(unittest.TestCase):
    def test_dumps(self):
        fixtures = [
            {"input": {}, "expected": "{}"},
            {"input": {"foo": "bar"}, "expected": '{"foo": "bar"}'},
            {"input": {"foo": b"bar"}, "expected": '{"foo": "bar"}'},
            {
                "input": {"foo": {"foo": {"foo": b"bar"}}},
                "expected": '{"foo": {"foo": {"foo": "bar"}}}',
            },
            {"input": [], "expected": "[]"},
            {"input": (), "expected": "[]"},
            {"input": [b"a", b"b", 1, 2.0], "expected": '["a", "b", 1, 2.0]'},
            {"input": {"foo": [b"bar"]}, "expected": '{"foo": ["bar"]}'},
            {
                "input": (
                    1,
                    2,
                    b"a",
                ),
                "expected": '[1, 2, "a"]',
            },
        ]
        for f in fixtures:
            self.assertEqual(keylime_json.dumps(f["input"]), f["expected"])



from keylime.common import exception

######################
# Keylime Agent States
######################

REGISTERED = 0
"""The agent is registered with registrar but not added to verifier yet"""

START = 1
"""The agent is added to verifier and will be moved to next state"""

SAVED = 2
"""The agent was added in verifier and wait for requests"""

GET_QUOTE = 3
"""The agent is under periodic integrity checking"""

GET_QUOTE_RETRY = 4
"""The agent is under periodic integrity checking but in a retry
state due to connection issues"""

PROVIDE_V = 5
"""The agent is receiving V key from the verifier"""

PROVIDE_V_RETRY = 6
"""The agent is recieving V key from the verifier but in a retry state due to
connection issues"""

FAILED = 7
"""The agent host failed to prove the integrity"""

TERMINATED = 8
"""The agent was terminated and will be removed from verifier"""

INVALID_QUOTE = 9
"""The integrity report from agent is not trusted against whitelist"""

TENANT_FAILED = 10
"""The agent was terminated but failed to be removed form verifier"""


VALID_STATES = (
    REGISTERED,
    START,
    SAVED,
    GET_QUOTE,
    GET_QUOTE_RETRY,
    PROVIDE_V,
    PROVIDE_V_RETRY,
    FAILED,
    TERMINATED,
    INVALID_QUOTE,
    TENANT_FAILED,
)

APPROVED_REACTIVATE_STATES = [START, GET_QUOTE, GET_QUOTE_RETRY, PROVIDE_V, PROVIDE_V_RETRY]

STATE_REPRESENTATIONS = {
    REGISTERED: "Registered",
    START: "Start",
    SAVED: "Saved",
    GET_QUOTE: "Get Quote",
    GET_QUOTE_RETRY: "Get Quote (retry)",
    PROVIDE_V: "Provide V",
    PROVIDE_V_RETRY: "Provide V (retry)",
    FAILED: "Failed",
    TERMINATED: "Terminated",
    INVALID_QUOTE: "Invalid Quote",
    TENANT_FAILED: "Tenant Quote Failed",
}


def state_to_str(state):
    if state not in VALID_STATES:
        raise exception.InvalidAgentState()
    return STATE_REPRESENTATIONS[state]



"""Store keyrings learned from IMA log

Revision ID: c3842cc9ee69
Revises: 257fe0f0c039
Create Date: 2021-09-14 13:12:29.306841

"""
import sqlalchemy as sa
from alembic import op

import keylime

# revision identifiers, used by Alembic.
revision = "c3842cc9ee69"
down_revision = "257fe0f0c039"
branch_labels = None
depends_on = None


def upgrade(engine_name):
    globals()[f"upgrade_{engine_name}"]()


def downgrade(engine_name):
    globals()[f"downgrade_{engine_name}"]()


def upgrade_registrar():
    pass


def downgrade_registrar():
    pass


def upgrade_cloud_verifier():
    op.add_column(
        "verifiermain", sa.Column("learned_ima_keyrings", keylime.db.verifier_db.JSONPickleType(), nullable=True)
    )


def downgrade_cloud_verifier():
    op.drop_column("verifiermain", "learned_ima_keyrings")



import argparse
import codecs
import itertools
import select
import sys
import time

from keylime.common import algorithms
from keylime.ima import ast
from keylime.tpm.tpm_abstract import config
from keylime.tpm.tpm_main import tpm

# Instaniate tpm
tpm_instance = tpm(need_hw_tpm=True)


def measure_list(file_path, position, ima_hash_alg, pcr_hash_alg, search_val=None):
    with open(file_path, encoding="utf-8") as f:
        lines = itertools.islice(f, position, None)

        runninghash = ast.get_START_HASH(pcr_hash_alg)

        if search_val is not None:
            search_val = codecs.decode(search_val.encode("utf-8"), "hex")

        for line in lines:
            # remove only the newline character, as there can be the
            # space as the delimiter character followed by an empty
            # field at the end
            line = line.strip("\n")
            position += 1

            entry = ast.Entry(line, None, ima_hash_alg=ima_hash_alg, pcr_hash_alg=pcr_hash_alg)

            if search_val is None:
                val = codecs.encode(entry.pcr_template_hash, "hex").decode("utf8")
                tpm_instance.extendPCR(config.IMA_PCR, val, pcr_hash_alg)
            else:
                runninghash = pcr_hash_alg.hash(runninghash + entry.pcr_template_hash)
                if runninghash == search_val:
                    return position

        if search_val is not None:
            raise Exception(
                "Unable to find current measurement list position, Resetting the TPM emulator may be neccesary"
            )

    return position


def main(argv=sys.argv):  # pylint: disable=dangerous-default-value
    parser = argparse.ArgumentParser()
    parser.add_argument("-a", "--hash_algs", nargs="*", default=["sha1"], help="PCR banks hash algorithms")
    parser.add_argument("-i", "--ima-hash-alg", default="sha1", help="Set hash algorithm that is used in IMA log")
    parser.add_argument("-f", "--ima-log", default=config.IMA_ML, help="path to the IMA log")
    args = parser.parse_args(argv[1:])

    if not tpm_instance.is_emulator():
        raise Exception("This stub should only be used with a TPM emulator")

    ima_hash_alg = algorithms.Hash(args.ima_hash_alg)
    position = {}
    for pcr_hash_alg in args.hash_algs:
        pcr_hash_alg = algorithms.Hash(pcr_hash_alg)
        position[pcr_hash_alg] = 0

    for pcr_hash_alg in dict(position):
        pcr_val = tpm_instance.readPCR(config.IMA_PCR, pcr_hash_alg)
        if codecs.decode(pcr_val.encode("utf-8"), "hex") != ast.get_START_HASH(pcr_hash_alg):
            print(
                f"Warning: IMA PCR is not empty for hash algorithm {pcr_hash_alg}, "
                "trying to find the last updated file in the measurement list..."
            )
            position[pcr_hash_alg] = measure_list(
                args.ima_log, position[pcr_hash_alg], ima_hash_alg, pcr_hash_alg, pcr_val
            )

    print(f"Monitoring {args.ima_log}")
    poll_object = select.poll()
    with open(args.ima_log, encoding="utf-8") as fd_object:
        number = fd_object.fileno()
        poll_object.register(fd_object, select.POLLIN | select.POLLPRI)

        try:
            while True:
                results = poll_object.poll()
                for result in results:
                    if result[0] != number:
                        continue
                    for pcr_hash_alg, pos in position.items():
                        position[pcr_hash_alg] = measure_list(args.ima_log, pos, ima_hash_alg, pcr_hash_alg)

                    time.sleep(0.2)
        except (SystemExit, KeyboardInterrupt):
            sys.exit(1)


if __name__ == "__main__":
    main(sys.argv)



from cryptography.hazmat.primitives.serialization import load_der_public_key
from pyasn1.codec.der import decoder, encoder
from pyasn1_modules import rfc2459


# Issue #944 -- python-cryptography won't parse malformed certs,
# such as some Nuvoton ones we have encountered in the field.
# Unfortunately, we still have to deal with such certs anyway.
# Let's read the EK cert with pyasn1 instead of python-cryptography.
def read_x509_der_cert_pubkey(der_cert_data):
    """Returns the public key of a DER-encoded X.509 certificate"""
    der509 = decoder.decode(der_cert_data, asn1Spec=rfc2459.Certificate())[0]
    return load_der_public_key(encoder.encode(der509["tbsCertificate"]["subjectPublicKeyInfo"]))



import tempfile

import gpg
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.serialization import load_pem_public_key

from keylime import keylime_logging

logger = keylime_logging.init_logging("signing")


def verify_signature_from_file(key_file, filename, sig_file, file_description):
    """
    Verify the file signature on disk (sig_file) using a public key on disk
    (key_file) with the file on disk (file). All inputs should be file
    paths.
    """

    with open(key_file, "rb") as key_f:
        key = key_f.read()
    with open(sig_file, "rb") as sig_f:
        sig = sig_f.read()
    with open(filename, "rb") as file_f:
        file = file_f.read()

    if verify_signature(key, sig, file):
        logger.debug("%s passed signature verification", file_description.capitalize())
    else:
        raise Exception(
            f"{file_description.capitalize()} signature verification failed comparing {file_description} ({filename}) against sig_file ({sig_file})"
        )


def verify_signature(key, sig, file):
    """
    Verify the file signature (sig) using a public key (key)
    with the file (file).
    """

    # Inspect the public key to determine what kind of key it is.
    key_header = key.decode("utf-8").split("\n")[0].strip()

    verified = False

    try:
        # PGP
        if key_header == "-----BEGIN PGP PUBLIC KEY BLOCK-----":
            verified = False
            with tempfile.TemporaryDirectory() as gpg_homedir:
                ctx = gpg.Context(home_dir=gpg_homedir)
                try:
                    logger.debug("Importing GPG key")
                    result = ctx.key_import(key)
                except Exception as e:
                    raise Exception("Unable to import GPG key") from e

                if result is not None and hasattr(result, "considered") is True:
                    _, result = ctx.verify(file, sig)
                    verified = result.signatures[0].status == 0

        # OpenSSL
        elif key_header == "-----BEGIN PUBLIC KEY-----":
            logger.debug("Importing ECDSA key")
            pubkey = load_pem_public_key(key)

            if isinstance(pubkey, ec.EllipticCurvePublicKey):
                logger.debug("EC public key successfully imported, verifying signature...")
                try:
                    pubkey.verify(sig, file, ec.ECDSA(hashes.SHA256()))
                    verified = True
                except InvalidSignature:
                    verified = False
            else:
                raise Exception(f"Unsupported public key algorithm: {type(pubkey)}")
        else:
            raise Exception("Unrecognized key type!")

    except Exception as e:
        logger.warning("Unable to verify signature: %s", e)
        verified = False

    return verified



"""
SPDX-License-Identifier: Apache-2.0
Copyright 2017 Massachusetts Institute of Technology.
"""

import sys
import unittest
from pathlib import Path

from keylime.crypto import (
    base64,
    decrypt,
    do_hmac,
    encrypt,
    generate_random_key,
    get_public_key,
    get_random_bytes,
    kdf,
    rsa_decrypt,
    rsa_encrypt,
    rsa_export_privkey,
    rsa_export_pubkey,
    rsa_generate,
    rsa_import_privkey,
    rsa_import_pubkey,
    rsa_sign,
    rsa_verify,
    strbitxor,
)

# Useful constants for the test
KEYLIME_DIR = Path(__file__).parents[1]

# Custom imports
sys.path.insert(0, KEYLIME_DIR)


class Crypto_Test(unittest.TestCase):
    def test_rsa(self):
        message = b"a secret message!"
        private = rsa_generate(2048)
        pubkeypem = rsa_export_pubkey(private)
        pubkey = rsa_import_pubkey(pubkeypem)
        keypem = rsa_export_privkey(private)
        key = rsa_import_privkey(keypem)
        ciphertext = rsa_encrypt(pubkey, message)
        plain = rsa_decrypt(key, ciphertext)
        self.assertEqual(plain, message)

    def test_aes(self):
        message = b"a secret message!"
        aeskey = get_random_bytes(32)
        ciphertext = encrypt(message, aeskey)
        plaintext = decrypt(ciphertext, aeskey)
        self.assertEqual(plaintext, message)

    def test_hmac(self):
        message = "a secret message!"
        aeskey = kdf(message, "salty-McSaltface")
        digest = do_hmac(aeskey, message)
        aeskey2 = kdf(message, "salty-McSaltface")
        self.assertEqual(do_hmac(aeskey2, message), digest)

    def test_xor(self):
        k = get_random_bytes(32)
        s1 = generate_random_key(32)
        s2 = strbitxor(s1, k)
        self.assertEqual(strbitxor(s1, s2), k)

    def test_errors(self):
        encrypt(None, get_random_bytes(32))

        with self.assertRaises(Exception):
            decrypt("", None)

        invalid = base64.b64encode(get_random_bytes(45))
        with self.assertRaises(Exception):
            decrypt(invalid, None)

    def test_rsa_sign(self):
        message = b"a secret message!"
        private = rsa_generate(2048)
        public_key = get_public_key(private)
        signature = rsa_sign(private, message)
        self.assertTrue(rsa_verify(public_key, message, signature))

        message = b"another message!"
        self.assertFalse(rsa_verify(public_key, message, signature))


if __name__ == "__main__":
    unittest.main()



import tempfile
import unittest

from keylime.ima import ima


class TestIMA(unittest.TestCase):
    def test_read_measurement_list(self):
        filedata = "0-entry\n1-entry\n2-entry\n"
        with tempfile.NamedTemporaryFile(delete=True) as tf:
            tf.write(filedata.encode("utf-8"))
            tf.flush()

            # Open file again to get str on read()
            with open(tf.name, "r", encoding="utf-8") as ima_log_file:

                # Request the 2nd entry, which is available
                ml, nth_entry, num_entries = ima.read_measurement_list(ima_log_file, 2)
                self.assertEqual(num_entries, 3)
                self.assertEqual(nth_entry, 2)
                self.assertTrue(ml.startswith("2-entry"))

                # Request the 3rd entry, which is not available yet, thus we get an empty list
                ml, nth_entry, num_entries = ima.read_measurement_list(ima_log_file, 3)
                self.assertEqual(num_entries, 3)
                self.assertEqual(nth_entry, 3)
                self.assertTrue(ml == "")

                # Request the 4th entry, which is beyond the next entry; since this is wrong,
                # we expect the entire list now.
                ml, nth_entry, num_entries = ima.read_measurement_list(ima_log_file, 4)
                self.assertEqual(num_entries, 3)
                self.assertEqual(nth_entry, 0)
                self.assertTrue(ml.startswith("0-entry"))


if __name__ == "__main__":
    unittest.main()



import base64
import enum
import json
import struct

from cryptography import utils, x509
from cryptography.exceptions import InvalidSignature, UnsupportedAlgorithm
from cryptography.hazmat import backends
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ec, padding
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from cryptography.hazmat.primitives.asymmetric.utils import Prehashed

from keylime import keylime_logging

logger = keylime_logging.init_logging("file_signatures")


"""
Tools for IMA file signature verification
"""


class HashAlgo(enum.IntEnum):
    """The hash_algo's as Linux defines them:
    https://elixir.bootlin.com/linux/latest/source/include/uapi/linux/hash_info.h#L17
    """

    HASH_ALGO_MD4 = 0
    HASH_ALGO_MD5 = 1
    HASH_ALGO_SHA1 = 2
    HASH_ALGO_RIPE_MD_160 = 3
    HASH_ALGO_SHA256 = 4
    HASH_ALGO_SHA384 = 5
    HASH_ALGO_SHA512 = 6
    HASH_ALGO_SHA224 = 7
    HASH_ALGO_RIPE_MD_128 = 8
    HASH_ALGO_RIPE_MD_256 = 9
    HASH_ALGO_RIPE_MD_320 = 10
    HASH_ALGO_WP_256 = 11
    HASH_ALGO_WP_384 = 12
    HASH_ALGO_WP_512 = 13
    HASH_ALGO_TGR_128 = 14
    HASH_ALGO_TGR_160 = 15
    HASH_ALGO_TGR_192 = 16
    HASH_ALGO_TGR_256 = 17
    HASH_ALGO_STREEBOG_256 = 18
    HASH_ALGO_STREEBOG_512 = 19


# Streebog is supported by evmctl
# pylint: disable=E1101
@utils.register_interface(hashes.HashAlgorithm)
class MyStreebog256:
    """Basic class for Streebog256"""

    name = "streebog256"
    digest_size = 32
    block_size = 64


# pylint: disable=E1101
@utils.register_interface(hashes.HashAlgorithm)
class MyStreebog512:
    """Basic class for Streebog512"""

    name = "streebog512"
    digest_size = 64
    block_size = 64


HASH_FUNCS = {
    # The list of hash functions we need for signature verification.
    HashAlgo.HASH_ALGO_MD5: hashes.__dict__.get("MD5"),
    HashAlgo.HASH_ALGO_SHA1: hashes.__dict__.get("SHA1"),
    HashAlgo.HASH_ALGO_SHA256: hashes.__dict__.get("SHA256"),
    HashAlgo.HASH_ALGO_SHA384: hashes.__dict__.get("SHA384"),
    HashAlgo.HASH_ALGO_SHA512: hashes.__dict__.get("SHA512"),
    HashAlgo.HASH_ALGO_SHA224: hashes.__dict__.get("SHA224"),
    HashAlgo.HASH_ALGO_STREEBOG_256: MyStreebog256,
    HashAlgo.HASH_ALGO_STREEBOG_512: MyStreebog512,
}


class EvmImaXattrType(enum.IntEnum):
    """https://elixir.bootlin.com/linux/v5.9.8/source/security/integrity/integrity.h#L74"""

    IMA_XATTR_DIGEST = 1
    EVM_XATTR_HMAC = 2
    EVM_IMA_XATTR_DIGSIG = 3
    IMA_XATTR_DIGEST_NG = 4
    EVM_XATTR_PORTABLE_DIGSIG = 5


class PubkeyAlgo(enum.IntEnum):
    """https://elixir.bootlin.com/linux/v5.9.8/source/include/linux/digsig.h#L17"""

    PUBKEY_ALGO_RSA = 0


class ImaKeyring:
    """ImaKeyring models an IMA keyring where keys are indexed by their keyid"""

    def __init__(self):
        """Constructor"""
        self.ringv2 = {}

    @staticmethod
    def _get_keyidv2(pubkey):
        """Calculate the keyidv2 of a given public key object. The keyidv2
        are the lowest 4 bytes of the sha1 hash over the public key bytes
        of a DER-encoded key in PKCS1 format.
        """
        if isinstance(pubkey, RSAPublicKey):
            fmt = serialization.PublicFormat.PKCS1
            pubbytes = pubkey.public_bytes(encoding=serialization.Encoding.DER, format=fmt)
        elif isinstance(pubkey, EllipticCurvePublicKey):
            fmt = serialization.PublicFormat.UncompressedPoint
            pubbytes = pubkey.public_bytes(encoding=serialization.Encoding.X962, format=fmt)
        else:
            raise UnsupportedAlgorithm(f"Unsupported public key type {type(pubkey)}")

        default_be = backends.default_backend()
        digest = hashes.Hash(hashes.SHA1(), backend=default_be)
        digest.update(pubbytes)
        keydigest = digest.finalize()
        return int.from_bytes(keydigest[16:], "big")

    def add_pubkey(self, pubkey, keyidv2):
        """Add a public key object to the keyring; a keyidv2 may be passed in
        and if it is 'None' it will be determined using the commonly used
        sha1 hash function for calculating the Subject Key Identifier.
        """
        if not keyidv2:
            keyidv2 = ImaKeyring._get_keyidv2(pubkey)
        # it's unlikely that two different public keys have the same 32 bit keyidv2
        self.ringv2[keyidv2] = pubkey
        logger.debug("Added key with keyid: 0x%08x", keyidv2)

    def get_pubkey_by_keyidv2(self, keyidv2):
        """Get a public key object given its keyidv2"""
        return self.ringv2.get(keyidv2)

    def to_json(self):
        """Convert the ImaKeyring into a JSON object"""
        fmt = serialization.PublicFormat.SubjectPublicKeyInfo
        obj = {}
        lst = []

        for pubkey in self.ringv2.values():
            try:
                pubbytes = pubkey.public_bytes(encoding=serialization.Encoding.DER, format=fmt)
            except Exception as ex:
                logger.error("Could not serialize key: %s", str(ex))
            lst.append(pubbytes)

        obj["pubkeys"] = [base64.b64encode(pubkey).decode("ascii") for pubkey in lst]
        obj["keyids"] = list(self.ringv2.keys())
        return obj

    def to_string(self):
        """Generate a string representation"""
        return json.dumps(self.to_json())

    @staticmethod
    def _base64_to_der_keylist(base64_keylist, keyidv2_list):
        """Convert a base64-encoded list of public keys to a list of DER-encoded
        public keys; a keyidv2_list may also be given that contains
        the keyidv2 of each key
        """
        res = []
        for idx, entry in enumerate(base64_keylist):
            keyidv2 = keyidv2_list[idx] if idx < len(keyidv2_list) else None
            res.append((base64.b64decode(entry), keyidv2))
        return res

    @staticmethod
    def from_string(stringrepr):
        """Convert a string-encoded ImaKeyring to an ImaKeyring object"""
        if not stringrepr:
            return None

        ima_keyring = ImaKeyring()

        default_be = backends.default_backend()

        # An empty Db entry comes as a string '[]'. A valid DB entry as a string
        # ith escaped quotes and needs to be loaded twice
        obj = json.loads(stringrepr)
        if isinstance(obj, str):
            obj = json.loads(obj)
        if not isinstance(obj, dict):
            return None

        keyids = obj.get("keyids", [])

        for (der_key, keyidv2) in ImaKeyring._base64_to_der_keylist(obj["pubkeys"], keyids):
            try:
                pubkey = serialization.load_der_public_key(der_key, backend=default_be)
                ima_keyring.add_pubkey(pubkey, keyidv2)
            except Exception as ex:
                logger.error("Could not load a base64-decoded DER key: %s", str(ex))
        return ima_keyring


class ImaKeyrings:
    """IMA Keyrings models the various keyrings of the system where IMA may take its keys from"""

    def __init__(self):
        """Constructor"""
        self.keyrings = {}

    def add_to_keyring_from_data(self, filedata, keyring_name):
        """Add the public key, given as a plain filedata (bytes), to the keyring by the given name
        after converting the key to an object. If a keyring by the given name doesn't exist, one
        will be created."""
        pubkey, keyidv2 = get_pubkey(filedata)
        self.add_pubkey_to_keyring(pubkey, keyring_name, keyidv2=keyidv2)

    def add_pubkey_to_keyring(self, pubkey, keyring_name, keyidv2=None):
        """Add a public key object to a keyring by the given name. If a keyring by the given name
        doesn't exist, one will be created."""
        keyring = self.keyrings.get(keyring_name)
        if not keyring:
            keyring = ImaKeyring()
            self.keyrings[keyring_name] = keyring
        keyring.add_pubkey(pubkey, keyidv2)

    def set_tenant_keyring(self, tenant_keyring):
        """Set the tenant keyring for which the tenant provided the keys via command line."""
        if tenant_keyring:
            self.keyrings["tenant_keyring"] = tenant_keyring
        else:
            self.keyrings.pop("tenant_keyring", None)

    def get_tenant_keyring(self):
        """Get the tenant keyring."""
        return self.keyrings.get("tenant_keyring")

    def get_all_keyrings(self):
        """Get a list of all the keyrings"""
        return list(self.keyrings.values())

    def to_json(self):
        """Convert an ImaKeyrings into its JSON representation; this does not include the tenant keyring"""
        obj = {}

        for name, keyring in self.keyrings.items():
            if name == "tenant_keyring":
                continue
            obj[name] = keyring.to_string()

        return obj

    def to_string(self):
        """Convert an ImaKeyrings into its string representation; this does not include the tenant keyring"""
        return json.dumps(self.to_json())

    @staticmethod
    def from_string(stringrepr):
        """Convert a string-encoded ImaKeyrings into an ImaKeyrings object."""

        obj = json.loads(stringrepr)
        if isinstance(obj, str):
            obj = json.loads(obj)
        if not isinstance(obj, dict):
            return None

        return ImaKeyrings.from_json(obj)

    @staticmethod
    def from_json(obj):
        """Convert a JSON representation of an ImaKeyrings object to an ImaKeyrings object."""
        ima_keyrings = ImaKeyrings()

        for name, keyring_str in obj.items():
            ima_keyrings.keyrings[name] = ImaKeyring.from_string(keyring_str)

        return ima_keyrings

    @staticmethod
    def _verify(pubkey, sig, filehash, hashfunc):
        """Do signature verification with the given public key"""
        if isinstance(pubkey, RSAPublicKey):
            pubkey.verify(sig, filehash, padding.PKCS1v15(), Prehashed(hashfunc))
        elif isinstance(pubkey, EllipticCurvePublicKey):
            pubkey.verify(sig, filehash, ec.ECDSA(Prehashed(hashfunc)))

    def _asymmetric_verify(self, signature, filehash, filehash_type):
        """Do an IMA signature verification given the signature data from
        the log, which is formatted as 'struct signature_v2_hdr'.
        This function resembles the kernel code:
        https://elixir.bootlin.com/linux/v5.9/source/security/integrity/digsig_asymmetric.c#L76
        https://elixir.bootlin.com/linux/v5.9/source/security/integrity/integrity.h#L116
        """

        siglen = len(signature)

        # The data are in big endian
        fmt = ">BBBIH"
        hdrlen = struct.calcsize(fmt)
        if len(signature) < hdrlen:
            logger.warning("Signature header is too short")
            return False
        _, _, hash_algo, keyidv2, sig_size = struct.unpack(fmt, signature[:hdrlen])

        siglen -= hdrlen

        if siglen != sig_size:
            logger.warning("Malformed signature")
            return False

        hashfunc = HASH_FUNCS.get(hash_algo)
        if not hashfunc:
            logger.warning("Unsupported hash algo with id '%d'", hash_algo)
            return False

        if filehash_type != hashfunc().name:
            logger.warning(
                "Mismatching filehash type %s and ima signature hash used %s", filehash_type, hashfunc().name
            )
            return False

        # Try all the keyrings until we find one with a key with the given keyidv2
        pubkey = None
        for keyring in self.get_all_keyrings():
            pubkey = keyring.get_pubkey_by_keyidv2(keyidv2)
            if pubkey:
                break

        if not pubkey:
            logger.warning("No key with id 0x%08x available", keyidv2)
            return False

        try:
            ImaKeyrings._verify(pubkey, signature[hdrlen:], filehash, hashfunc())
        except InvalidSignature:
            return False
        return True

    def integrity_digsig_verify(self, signature, filehash, filehash_type):
        """Validate the signature against the given hash trying all keyrings.
        This function resembles the kernel code at:
        https://elixir.bootlin.com/linux/v5.9/source/security/integrity/digsig.c#L59
        """
        fmt = ">BB"
        if len(signature) < struct.calcsize(fmt):
            logger.warning("Malformed signature: not enough bytes")
            return False

        typ, version = struct.unpack(fmt, signature[: struct.calcsize(fmt)])
        if typ not in [EvmImaXattrType.EVM_IMA_XATTR_DIGSIG, EvmImaXattrType.EVM_XATTR_PORTABLE_DIGSIG]:
            logger.warning("Malformed signature: wrong type")
            return False

        if version == 2:
            return self._asymmetric_verify(signature, filehash, filehash_type)

        logger.warning("Malformed signature: wrong version (%d)", version)
        return False


def _get_pubkey_from_der_public_key(filedata, backend):
    """Load the filedata as a DER public key"""
    try:
        return serialization.load_der_public_key(filedata, backend=backend), None
    except Exception:
        return None, None


def _get_pubkey_from_pem_public_key(filedata, backend):
    """Load the filedata as a PEM public key"""
    try:
        return serialization.load_pem_public_key(filedata, backend=backend), None
    except Exception:
        return None, None


def _get_pubkey_from_der_private_key(filedata, backend):
    """Load the filedata as a DER private key"""
    try:
        privkey = serialization.load_der_private_key(filedata, None, backend=backend)
        return privkey.public_key(), None
    except Exception:
        return None, None


def _get_pubkey_from_pem_private_key(filedata, backend):
    """Load the filedata as a PEM private key"""
    try:
        privkey = serialization.load_pem_private_key(filedata, None, backend=backend)
        return privkey.public_key(), None
    except Exception:
        return None, None


def _get_keyidv2_from_cert(cert):
    """Get the keyidv2 from the cert's Subject Key Identifier"""
    if not cert.extensions:
        return None

    skid = cert.extensions.get_extension_for_oid(x509.oid.ExtensionOID.SUBJECT_KEY_IDENTIFIER)
    if skid and skid.value and len(skid.value.digest) >= 4:
        keyidv2 = int.from_bytes(skid.value.digest[-4:], "big")
        logger.debug("Extracted keyidv2 from cert: 0x%08x", keyidv2)
        return keyidv2
    return None


def _get_pubkey_from_der_x509_certificate(filedata, backend):
    """Load the filedata as a DER x509 certificate"""
    try:
        cert = x509.load_der_x509_certificate(filedata, backend=backend)
        return cert.public_key(), _get_keyidv2_from_cert(cert)
    except Exception:
        return None, None


def _get_pubkey_from_pem_x509_certificate(filedata, backend):
    """Load the filedata as a PEM x509 certificate"""
    try:
        cert = x509.load_pem_x509_certificate(filedata, backend=backend)
        return cert.public_key(), _get_keyidv2_from_cert(cert)
    except Exception:
        return None, None


def get_pubkey(filedata):
    """Get the public key from the filedata; if an x509 certificate is
    given, also determine the keyidv2 from the Subject Key Identifier,
    otherwise return None
    To make it easy for the user, we try to parse the filedata as
    PEM- or DER-encoded public key, x509 certificate, or even private key.
    This function then returns the public key object or None if the file
    contents could not be interpreted as a key.
    """
    default_be = backends.default_backend()
    for func in [
        _get_pubkey_from_der_x509_certificate,
        _get_pubkey_from_pem_x509_certificate,
        _get_pubkey_from_der_public_key,
        _get_pubkey_from_pem_public_key,
        _get_pubkey_from_der_private_key,
        _get_pubkey_from_pem_private_key,
    ]:
        pubkey, keyidv2 = func(filedata, default_be)
        if pubkey:
            return pubkey, keyidv2

    return None, None


def get_pubkey_from_file(filename):
    """Get the public key object from a file"""
    try:
        with open(filename, "rb") as fobj:
            filedata = fobj.read()
            pubkey, keyidv2 = get_pubkey(filedata)
            if pubkey:
                return pubkey, keyidv2
    except Exception:
        pass

    return None, None



"""Tagging of failure events that might cause revocation in Keylime.

"""

import enum
import functools
import re
from typing import Any, Callable, Dict, List, Optional, Pattern, Tuple, Union

from keylime import json, keylime_logging

logger = keylime_logging.init_logging("failure")


@functools.total_ordering
class SeverityLabel:
    """
    Severity label that can be attached to an event.

    The severity level is assigned dynamically based on the configuration,
    so only use the name for use outside use of the tagging module.
    """

    name: str
    severity: int

    def __init__(self, name: str, severity: int):
        self.name = name
        self.severity = severity

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, SeverityLabel):
            return NotImplemented
        return self.severity < other.severity

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SeverityLabel):
            return NotImplemented
        return self.severity == other.severity


class Component(enum.Enum):
    """
    Main components of Keylime that can generate revocations.
    """

    QUOTE_VALIDATION = "qoute_validation"
    PCR_VALIDATION = "pcr_validation"
    MEASURED_BOOT = "measured_boot"
    IMA = "ima"
    INTERNAL = "internal"
    DEFAULT = "default"


class Event:
    """
    Event that might be the reason for revocation.

    The context is string
    """

    event_id: str
    severity_label: SeverityLabel
    context: str
    recoverable: bool

    def __init__(
        self,
        component: Component,
        sub_components: Optional[List[str]],
        event_id: str,
        context: Union[str, Dict[str, json.JSONType]],
        recoverable: bool,
    ):

        # Build full event id with the format "component.[sub_component].event_id"
        self.event_id = component.value
        if sub_components is not None:
            self.event_id += "." + ".".join(sub_components)
        self.event_id += f".{event_id}"

        # Convert message
        if isinstance(context, str):
            context = {"message": context}
        self.context = json.dumps(context)

        self.severity_label = _severity_match(self.event_id)
        self.recoverable = recoverable


class Failure:
    """
    Failure Object that collects all events that might cause a revocation.

    If recoverable is set to False the validation process returned early and might skipped other validation steps.
    """

    events: List[Event]
    recoverable: bool
    highest_severity: Optional[SeverityLabel]
    _component: Component
    _sub_components: Optional[List[str]]

    def __init__(self, component: Component, sub_components: Optional[List[str]] = None):
        self._component = component
        self._sub_components = sub_components
        self.events = []
        self.recoverable = True
        self.highest_severity_event: Optional[
            Event
        ] = None  # This only holds the first event that has the highest severity
        self.highest_severity: Optional[SeverityLabel] = None

    def _add(self, event: Event) -> None:
        if not event.recoverable:
            self.recoverable = False
            if event.severity_label != MAX_SEVERITY_LABEL:
                logger.warning(
                    "Irrecoverable Event with id: %s has not the highest severity level.\n "
                    "Setting it the the highest severity level.",
                    event.event_id,
                )
                event.severity_label = MAX_SEVERITY_LABEL

        if self.highest_severity is None or event.severity_label > self.highest_severity:
            self.highest_severity = event.severity_label
            self.highest_severity_event = event

        self.events.append(event)

    def add_event(
        self,
        event_id: str,
        context: Union[str, Dict[str, json.JSONType]],
        recoverable: bool,
        sub_components: Optional[List[str]] = None,
    ) -> None:
        """
        Add event to Failure object. Uses the component and subcomponents specified in the Failure object.

        As context specify either a string that contains a message or a dict that contains useful information about that
        event.
        Set recoverable to False if the code skips other not directly related checks. Those events should always have
        the highest severity label assigned and if not we manually do that.

        Example usage:
            failure.add_event("ima_hash",
                              {"message": "IMA hash does not match the calculated hash.",
                               "expected": self.template_hash, "got": self.mode.hash()}, True)
        """
        if sub_components is not None and self._sub_components is not None:
            sub_components = self._sub_components + sub_components
        elif self._sub_components is not None:
            sub_components = self._sub_components
        event = Event(self._component, sub_components, event_id, context, recoverable)
        self._add(event)

    def merge(self, other: "Failure") -> None:
        if self.recoverable:
            self.recoverable = other.recoverable
        if self.highest_severity is None:
            self.highest_severity = other.highest_severity
            self.highest_severity_event = other.highest_severity_event
        elif other.highest_severity is not None and self.highest_severity < other.highest_severity:
            self.highest_severity = other.highest_severity
            self.highest_severity_event = other.highest_severity_event

        self.events.extend(other.events)

    def get_event_ids(self) -> List[str]:
        """
        Returns a list of the event ids currently collected
        """
        return list(map(lambda x: x.event_id, self.events))

    def __bool__(self) -> bool:
        return not self.recoverable or len(self.events) > 0


def _eval_severity_config(
    severity_labels=None, severity_policy=None
) -> Tuple[List[Callable[[str], Optional[SeverityLabel]]], SeverityLabel]:
    """
    Generates the list of rules to match a event_id against.
    """

    if severity_labels:
        labels_list = severity_labels
    else:
        # Default labels from lowest to highest severity
        labels_list = ["info", "notice", "warning", "error", "critical", "alert", "emergency"]

    if severity_policy:
        policies = severity_policy
    else:
        # The default policy assigns the highest severity to all events
        policies = [{"event_id": ".*", "severity_label": labels_list[-1]}]

    labels = {}
    for label, level in zip(labels_list, range(0, len(labels_list))):
        labels[label] = SeverityLabel(label, level)

    label_max = labels[labels_list[-1]]

    rules = []
    for policy in policies:
        # TODO validate regex
        regex = re.compile(policy["event_id"])

        def rule(policy_regex: Pattern[Any], label_str: str, event_id: str) -> Optional[SeverityLabel]:
            if policy_regex.fullmatch(event_id):
                policy_label = labels.get(label_str)
                if policy_label is None:
                    logger.error("Label %s is not a valid label. Defaulting to maximal severity label!", label_str)
                    return label_max
                return policy_label
            return None

        new_rule: Callable[[str], Optional[SeverityLabel]] = functools.partial(rule, regex, policy["severity_label"])
        rules.append(new_rule)
    return rules, label_max


# Only evaluate the policy once on module load
SEVERITY_RULES, MAX_SEVERITY_LABEL = _eval_severity_config()


def set_severity_config(severity_labels, severity_policy):
    """
    Set a custom severity configuration
    """

    if not severity_labels:
        raise Exception("Empty severity_labels provided to set_severity_config()")

    if not severity_policy:
        raise Exception("Empty severity_policy provided to set_severity_config()")

    global SEVERITY_RULES, MAX_SEVERITY_LABEL

    SEVERITY_RULES, MAX_SEVERITY_LABEL = _eval_severity_config(
        severity_labels=severity_labels, severity_policy=severity_policy
    )

    logger.info("Severity configuration set")


def _severity_match(event_id: str) -> SeverityLabel:
    """
    Match the event_id to a severity label.
    """
    for rule in SEVERITY_RULES:
        match = rule(event_id)
        if match is not None:
            return match
    logger.warning("No rule matched for event_id: %s. Defaulting to max severity label", event_id)
    return MAX_SEVERITY_LABEL



class KeylimeException(Exception):
    """Base class for all Keylime exceptions"""

    _msg_fmt = "An unknown exception occurred."

    def __init__(self, message=None, **kwargs):
        if not message:
            message = self._msg_fmt % kwargs

        super().__init__(message)


class InvalidAgentState(KeylimeException):
    _msg_fmt = "Invalid agent state."



"""Parser and validator for device mapper IMA events

  - https://www.kernel.org/doc/html/v5.15/admin-guide/device-mapper/dm-ima.html

"""

import pickle
import re
import sys
from collections import ChainMap
from itertools import chain
from typing import Dict, List, Optional, Union

import lark
from lark import Lark, Transformer, v_args
from packaging import version

from keylime.common.algorithms import Hash
from keylime.failure import Component, Failure
from keylime.ima import ast
from keylime.ima.dm_grammar import DM_GRAMMAR

if sys.version_info >= (3, 7):
    from dataclasses import dataclass
else:
    from keylime.backport_dataclasses import dataclass


class DeepChainMap(ChainMap):
    """
    Variant of ChainMap that handles updates for nested ChainMaps
    Based on the example in https://docs.python.org/3/library/collections.html#collections.ChainMap
    """

    def __setitem__(self, key, value):
        for mapping in self.maps:
            if key in mapping:
                mapping[key] = value
                return
        super().__setitem__(key, value)

    def __delitem__(self, key):
        for mapping in self.maps:
            if key in mapping:
                del mapping[key]
                return
        raise KeyError(key)


@dataclass
class DeviceState:
    policy_name: str
    active: bool
    valid_state: bool
    active_table_hash: Optional[ast.Digest]
    inactive_table_hash: Optional[ast.Digest]
    num_targets: int
    allow_multiple_loads: bool


class DmIMAValidator:
    valid_names = [
        "dm_table_load",
        "dm_device_resume",
        "dm_device_remove",
        "dm_table_clear",
        "dm_device_rename",
        "dm_target_update",
    ]
    policies: dict
    devices: Dict[str, DeviceState]

    def __init__(self, policies) -> None:
        self.policies = policies
        self.devices = {}

    def state_load(self, state: Optional[bytes]) -> None:
        if state:
            self.devices = pickle.loads(state)

    def state_dump(self) -> bytes:
        """
        Export the state of the validator.
        Note: the validator assumes that this data is always trustworthy and might not compatible between versions
        """
        return pickle.dumps(self.devices)

    def validate(self, digest: ast.Digest, path: ast.Name, data: ast.Buffer) -> Failure:
        """Validate a single entry."""
        failure = Failure(Component.IMA, ["validation", "dm"])
        try:
            event = parse(data.data.decode("utf-8"), path.name)
            hash_alg = Hash(digest.algorithm)
            if digest.hash != hash_alg.hash(data.data):
                failure.add_event("invalid_data", "hash in IMA log and of the actual data mismatch", True)

            match_key = self.policies["match_on"]

            if path.name == "dm_table_load":
                failure.merge(self.validate_table_load(event, match_key, digest))
            elif path.name == "dm_device_resume":
                failure.merge(self.validate_device_resume(event, match_key))
            elif path.name == "dm_device_remove":
                failure.merge(self.validate_device_remove(event, match_key))
            elif path.name == "dm_device_rename":
                failure.merge(self.validate_device_rename(event, match_key))
            elif path.name == "dm_table_clear":
                failure.merge(self.validate_table_clear(event, match_key))
            elif path.name == "dm_target_update":
                failure.merge(self.validate_target_update(event, match_key))
            else:
                failure.add_event("invalid_event_type", {"got": path.name}, True)

        except lark.exceptions.LarkError as e:
            failure.add_event("parsing_failed", f"Could not construct valid entry: {e}", True)

        return failure

    def validate_table_clear(self, event: "ClearEvent", match_key) -> Failure:
        failure = Failure(Component.IMA, ["validation", "dm", "dm_table_clear"])

        device_key = getattr(event.device_metadata, match_key)
        device_state = self.devices.get(device_key, None)
        if device_state is None:
            failure.add_event("clear_before_table_load", "Clear event before table was loaded", True)
            return failure

        policy = self.policies["rules"][device_state.policy_name]
        if not policy["allow_clear"]:
            failure.add_event(
                "table_cleared", f"Table for device {device_key} was cleared, but that is not allowed", True
            )
            device_state.valid_state = False
            return failure

        return failure

    def validate_target_update(self, event: "UpdateEvent", match_key) -> Failure:
        failure = Failure(Component.IMA, ["validation", "dm", "dm_target_update"])

        device_key = getattr(event.device_metadata, match_key)
        device_state = self.devices.get(device_key, None)
        if device_state is None:
            failure.add_event("update_before_table_load", "Update event before table was loaded", True)
            return failure

        policy = self.policies["rules"][device_state.policy_name]
        target_data = policy["table_load"]["targets"][event.target.target_index]

        # Validate the target data again
        self.validate_target_table(event.target, target_data, failure)

        return failure

    def validate_device_remove(self, event: "RemoveEvent", match_key) -> Failure:
        failure = Failure(Component.IMA, ["validation", "dm", "dm_device_remove"])

        # TODO: check if we can always use the active table
        device_key = getattr(event.device_active_metadata, match_key)
        device_state = self.devices.get(device_key, None)
        if device_state is None:
            failure.add_event("remove_before_table_load", "Remove event before table was loaded", True)
            return failure

        policy = self.policies["rules"][device_state.policy_name]
        if not policy["device_remove"]["allow_removal"]:
            failure.add_event("device_removed", f"Device {device_key} was remove, but that is not allowed", True)
            device_state.valid_state = False
            return failure

        # Remove device completely
        del self.devices[device_key]
        return failure

    def validate_device_resume(self, event: "ResumeEvent", match_key) -> Failure:
        failure = Failure(Component.IMA, ["validation", "dm", "dm_device_resume"])

        device_key = getattr(event.device_metadata, match_key)
        if device_key not in self.devices:
            failure.add_event("resume_before_table_load", "Resume event before table was loaded", True)
            return failure

        device_state = self.devices[device_key]

        # We only expect one resume event
        if device_state.active:
            failure.add_event("already_active", "table is already active", True)
            device_state.valid_state = False
            return failure

        # Check if the table hash is consistent
        if device_state.active_table_hash != event.active_table_hash:
            failure.add_event(
                "active_table_mismatch",
                {
                    "got": event.active_table_hash,
                    "expected": device_state.active_table_hash,
                    "context": "resume does not match the table",
                },
                True,
            )
            device_state.valid_state = False
            return failure

        # TODO: Check also current capacity.

        # Mark now device as active and in a valid state
        device_state.active = True
        device_state.valid_state = True
        return failure

    def validate_device_rename(self, event: "RenameEvent", match_key) -> Failure:
        failure = Failure(Component.IMA, ["validation", "dm", "dm_device_rename"])

        device_key = getattr(event.device_metadata, match_key)
        device_state = self.devices.get(device_key, None)
        if device_state is None:
            failure.add_event("rename_before_table_load", "Rename event before table was loaded", True)
            return failure

        policy = self.policies["rules"][device_state.policy_name]

        # Only check if the name changed
        if event.device_metadata.name != event.new_name:
            if not _check_attr(event.new_name, policy["device_rename"]["valid_name"]):
                failure.add_event(
                    "new_name_invalid",
                    {
                        "message": "New name is invalid",
                        "got": event.new_name,
                        "expected": policy["device_rename"]["valid_name"],
                    },
                    True,
                )
                device_state.valid_state = False
            elif match_key == "name":
                self.devices[event.new_name] = self.devices.pop(device_key)

        # Only check if uuid changed
        if event.device_metadata.uuid != event.new_uuid:
            if not _check_attr(event.new_uuid, policy["device_rename"]["valid_uuid"]):
                failure.add_event(
                    "new_uuid_invalid",
                    {
                        "message": "New name is invalid",
                        "got": event.new_uuid,
                        "expected": policy["device_rename"]["valid_uuid"],
                    },
                    True,
                )
                device_state.valid_state = False
            elif match_key == "uuid":
                self.devices[event.new_name] = self.devices.pop(device_key)

        return failure

    def validate_table_load(self, event: "LoadEvent", match_key, digest) -> Failure:
        failure = Failure(Component.IMA, ["validation", "dm", "dm_table_load"])

        device_key = getattr(event.device_metadata, match_key)

        if device_key in self.devices and not self.devices[device_key].allow_multiple_loads:
            failure.add_event("multiple_table_loads", f"Multiple table load entries for device: {device_key}", True)
            return failure

        # Find matching policy
        used_policy_name = None
        used_policy = None
        for policy_name, policy in self.policies["rules"].items():
            if re.fullmatch(policy["table_load"][match_key], device_key):
                used_policy = policy
                used_policy_name = policy_name
                break

        if used_policy_name is None:
            failure.add_event("no_matching_policy", "No policy found", True)
            return failure

        # Validate device metadata
        for entry in ["name", "uuid", "major", "minor", "minor_count", "num_targets"]:
            if not _check_attr(getattr(event.device_metadata, entry), used_policy["table_load"][entry]):
                failure.add_event(
                    "invalid_entry",
                    {
                        "got": getattr(event.device_metadata, entry),
                        "expected": used_policy["table_load"][entry],
                        "context": entry,
                    },
                    True,
                )
                return failure

        # Check "num_targets"
        # Note that we get actually could get multiple lines, but this does not happen for our use cases so we
        # treat it as a failure.
        if event.device_metadata.num_targets != len(event.targets) or used_policy["table_load"][entry] != len(
            event.targets
        ):
            failure.add_event("num_targets_mismatch", "lengths are not consistent", True)
            return failure

        # Validate targets
        for actual_target, should_values in zip(event.targets, used_policy["table_load"]["targets"]):
            self.validate_target_table(actual_target, should_values, failure)

        # Add device to validator
        self.devices[device_key] = DeviceState(
            active=False,
            valid_state=not used_policy["device_resume_required"],
            # Device normally has a resume to be in a fully validated state
            active_table_hash=digest,
            inactive_table_hash=None,
            policy_name=used_policy_name,
            allow_multiple_loads=used_policy["table_load"]["allow_multiple_loads"],
            num_targets=event.device_metadata.num_targets,
        )
        return failure

    @staticmethod
    def validate_target_table(target, reference_values, failure: Failure) -> None:
        """
        Validates a target table entry against reference_values in a policy.
        If a failure occurs it is added to the failure object.
        """
        valid = True
        for key in reference_values.keys():
            data = target
            # Non default target arguments are stored in target_attributes
            if key not in ["target_index", "target_begin", "target_len", "target_name", "target_version"]:
                data = target.target_attributes
            try:
                if not _check_attr(getattr(data, key), reference_values[key]):
                    failure.add_event(
                        "target_data_mismatch",
                        {"got": getattr(data, key), "expected": reference_values[key], "context": key},
                        True,
                    )
                    valid = False
            except AttributeError as e:
                failure.add_event(
                    "target_attribute_not_found", {"context": f"Key {key} not found on target: {e}"}, True
                )

        if not valid:
            failure.add_event("target_data_invalid", "target data was not valid", True)

    def invalid(self) -> Failure:
        """
        Check if the devices are in a consistent state
        """
        failure = Failure(Component.IMA, ["validation", "dm"])
        used_policies = set()
        # Check if the devices are in a valid state
        for device_name, device in self.devices.items():
            if not device.valid_state:
                failure.add_event("device_invalid_state", {"context": device_name}, True)
            else:
                # Only add polices that are on valid states
                used_policies.add(device.policy_name)

        # Check for required active policies
        for policy_name, policy in self.policies["rules"].items():
            if policy["required"] and policy_name not in used_policies:
                failure.add_event("required_policy_not_in_use", f"policy {policy_name} required but not used", True)
        return failure


def _strtobool(val: str) -> bool:
    """Convert a string representation of truth to true (1) or false (0).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return True
    if val in ("n", "no", "f", "false", "off", "0"):
        return False

    raise ValueError(f"invalid truth value {val}")


def _check_attr(attr: Optional[Union[int, str, bool]], reference_value: Optional[Union[int, str, bool]]) -> bool:
    """
    Validate an attribute against the reference value
    - If the reference value is a str we assume that is a regex
    - If the reference value it is a bool or int it is checked for equality after converting attr to the type
    - If the reference value is None we always return True
    - If the reference value is not None and the attr is or type conversion fails it returns False
    """
    if reference_value is None:
        return True

    if attr is None and reference_value is not None:
        return False

    if isinstance(reference_value, bool):

        if isinstance(attr, str):
            try:
                return reference_value == _strtobool(attr)
            except ValueError:
                return False
        else:
            return reference_value == attr

    # This has to be done after the bool check because bool is also an instance of int
    if isinstance(reference_value, int):
        try:
            return reference_value == int(attr)
        except ValueError:
            return False

    if isinstance(reference_value, str):
        return bool(re.fullmatch(reference_value, str(attr)))

    return False


class TargetAttributes:
    pass


@dataclass
class VerityAttributes(TargetAttributes):
    hash_failed: str
    verity_version: int
    data_device_name: str
    hash_device_name: str
    verity_algorithm: str
    root_digest: str
    salt: str
    ignore_zero_blocks: str
    check_at_most_once: str
    root_hash_sig_key_desc: Optional[str] = None
    verity_mode: Optional[str] = None


@dataclass
class CacheAttributes(TargetAttributes):
    metadata_mode: str
    cache_metadata_device: str
    cache_device: str
    cache_origin_device: str
    writethrough: str
    writeback: str
    passthrough: str
    metadata2: str
    no_discard_passdown: str


@dataclass
class CryptAttributes(TargetAttributes):
    allow_discards: str
    same_cpu_crypt: str
    submit_from_crypt_cpus: str
    no_read_workqueue: str
    no_write_workqueue: str
    iv_large_sectors: str
    key_size: int
    key_parts: int
    key_extra_size: int
    key_mac_size: int
    integrity_tag_size: Optional[int] = None
    cipher_auth: Optional[str] = None
    sector_size: Optional[int] = None
    cipher_string: Optional[str] = None


@dataclass
class IntegrityAttributes(TargetAttributes):
    dev_name: str
    start: int
    tag_size: int
    mode: str
    recalculate: str
    allow_discards: str
    fix_padding: str
    fix_hmac: str
    legacy_recalculate: str
    journal_sectors: int
    interleave_sectors: int
    buffer_sectors: int
    meta_device: Optional[str] = None
    block_size: Optional[int] = None


@dataclass
class LinearAttributes(TargetAttributes):
    device_name: str
    start: str


@dataclass
class SnapshotAttributes(TargetAttributes):
    snap_origin_name: str
    snap_cow_name: str
    snap_valid: str
    snap_merge_failed: str
    snapshot_overflowed: str


@dataclass
class MirrorDevice:
    """
    Device information used by the mirror target attributes
    """

    mirror_device_name: str  # In raw data this is "mirror_device_X"
    mirror_device_status: str  # In raw data this is "mirror_device_X_status"


@dataclass
class MirrorAttributes(TargetAttributes):
    nr_mirrors: str
    mirror_device_data: List[MirrorDevice]
    handle_errors: str
    keep_log: str
    log_type_status: str


@dataclass
class Target:
    target_index: int
    target_begin: int
    target_len: int
    target_name: str
    target_version: str
    target_attributes: TargetAttributes


@dataclass
class DeviceMetaData:
    """
    Generic device metadata.
    """

    name: str
    uuid: str
    major: int
    minor: int
    minor_count: int
    num_targets: int


@dataclass
class DeviceMetaDataMinimal:
    """
    Device metadata that is measured when no_data happens
    """

    name: str
    uuid: str


@dataclass
class LoadEvent:
    dm_version: str
    device_metadata: DeviceMetaData
    targets: List[Target]


@dataclass
class ResumeEvent:
    dm_version: str
    device_metadata: DeviceMetaData
    current_device_capacity: int
    active_table_hash: ast.Digest

    def __post_init__(self):
        if isinstance(self.active_table_hash, str):
            self.active_table_hash = ast.Digest(self.active_table_hash)


@dataclass
class RemoveEvent:
    dm_version: str
    remove_all: str
    current_device_capacity: int
    device_active_metadata: Optional[DeviceMetaData] = None
    device_inactive_metadata: Optional[DeviceMetaData] = None
    active_table_hash: Optional[str] = None
    inactive_table_hash: Optional[str] = None


@dataclass
class ClearEvent:
    dm_version: str
    current_device_capacity: int
    no_data: bool
    device_metadata: Optional[Union[DeviceMetaData, DeviceMetaDataMinimal]] = None
    inactive_table_hash: Optional[str] = None


@dataclass
class RenameEvent:
    dm_version: str
    device_metadata: DeviceMetaData  # This is the device metadata of the active one
    new_name: str
    new_uuid: str
    current_device_capacity: int


@dataclass
class UpdateEvent:
    dm_version: str
    device_metadata: DeviceMetaData
    target: Target


def _token_to_dict(name, prefix=None):
    """
    Generates function that converts a token to a dict containing the token name as key.
    The length of prefix is stripped from the key.
    """
    if prefix is not None:
        return lambda x: {name[len(prefix) :]: x[0]}
    return lambda x: {name: x[0]}


class DeviceMapperTransformer(Transformer):
    """
    Converts the Lark AST into the data structures for validation.
    """

    # All the tokens specified here just need to converted to a dict and in some cases their prefix stripped
    verity_tokens = [
        "verity_hash_failed",
        "verity_verity_version",
        "verity_data_device_name",
        "verity_hash_device_name",
        "verity_verity_algorithm",
        "verity_root_digest",
        "verity_salt",
        "verity_ignore_zero_blocks",
        "verity_check_at_most_once",
        "verity_root_hash_sig_key_desc",
        "verity_verity_mode",
    ]
    cache_tokens = [
        "cache_metadata_mode",
        "cache_cache_metadata_device",
        "cache_cache_device",
        "cache_cache_origin_device",
        "cache_writethrough",
        "cache_writeback",
        "cache_passthrough",
        "cache_metadata2",
        "cache_no_discard_passdown",
    ]
    crypt_tokens = [
        "crypt_allow_discards",
        "crypt_same_cpu_crypt",
        "crypt_submit_from_crypt_cpus",
        "crypt_no_read_workqueue",
        "crypt_no_write_workqueue",
        "crypt_iv_large_sectors",
        "crypt_integrity_tag_size",
        "crypt_cipher_auth",
        "crypt_sector_size",
        "crypt_cipher_string",
        "crypt_key_size",
        "crypt_key_parts",
        "crypt_key_extra_size",
        "crypt_key_mac_size",
    ]
    integrity_tokens = [
        "integrity_dev_name",
        "integrity_start",
        "integrity_tag_size",
        "integrity_mode",
        "integrity_meta_device",
        "integrity_block_size",
        "integrity_recalculate",
        "integrity_allow_discards",
        "integrity_fix_padding",
        "integrity_fix_hmac",
        "integrity_legacy_recalculate",
        "integrity_journal_sectors",
        "integrity_interleave_sectors",
        "integrity_buffer_sectors",
    ]
    mirror_tokens = ["mirror_nr_mirrors", "mirror_handle_errors", "mirror_keep_log", "mirror_log_type_status"]
    linear_tokens = ["linear_device_name", "linear_start"]
    snapshot_tokens = [
        "snapshot_snap_origin_name",
        "snapshot_snap_cow_name",
        "snapshot_snap_valid",
        "snapshot_snap_merge_failed",
        "snapshot_snapshot_overflowed",
    ]
    target_tokens = ["target_index", "target_begin", "target_len", "target_name", "target_version", "target_attributes"]
    device_tokens = [
        "device_name",
        "device_uuid",
        "device_major",
        "device_minor",
        "device_minor_count",
        "device_num_targets",
    ]
    other_tokens = ["dm_version", "remove_all", "current_device_capacity", "active_table_hash", "inactive_table_hash"]
    rename_tokens = ["rename_new_name", "rename_new_uuid"]

    def __init__(self, *args, **kwargs):
        for token in chain(self.target_tokens, self.other_tokens):
            setattr(self, token, _token_to_dict(token))

        for token in self.device_tokens:
            setattr(self, token, _token_to_dict(token, "device_"))

        for token in self.rename_tokens:
            setattr(self, token, _token_to_dict(token, "rename_"))

        # Target specific tokens
        for token in self.verity_tokens:
            setattr(self, token, _token_to_dict(token, "verity_"))
        for token in self.cache_tokens:
            setattr(self, token, _token_to_dict(token, "cache_"))
        for token in self.crypt_tokens:
            setattr(self, token, _token_to_dict(token, "crypt_"))
        for token in self.integrity_tokens:
            setattr(self, token, _token_to_dict(token, "integrity_"))
        for token in self.mirror_tokens:
            setattr(self, token, _token_to_dict(token, "mirror_"))
        for token in self.linear_tokens:
            setattr(self, token, _token_to_dict(token, "linear_"))
        for token in self.snapshot_tokens:
            setattr(self, token, _token_to_dict(token, "snapshot_"))

        super().__init__(*args, **kwargs)

    @staticmethod
    def INT(tok):
        return int(tok.value)

    @staticmethod
    def NUMBER(tok):
        return int(tok.value)

    @staticmethod
    def STRING(tok):
        return str(tok.value)

    @staticmethod
    def optional_string(children):
        """Special handling for options that can have an empty string"""
        if not children:
            return ""
        return children[0]

    @staticmethod
    @v_args(inline=True)
    def version_nb(a, b, c):
        return f"{a}.{b}.{c}"

    @staticmethod
    def yes(x):
        assert x == []
        return "y"

    @staticmethod
    def no(x):
        assert x == []
        return "n"

    @staticmethod
    def verity_attributes(children):
        return VerityAttributes(**DeepChainMap(*children))

    @staticmethod
    def cache_attributes(children):
        return CacheAttributes(**DeepChainMap(*children))

    @staticmethod
    def crypt_attributes(children):
        return CryptAttributes(**DeepChainMap(*children))

    @staticmethod
    def integrity_attributes(children):
        return IntegrityAttributes(**DeepChainMap(*children))

    @staticmethod
    def linear_attributes(children):
        return LinearAttributes(**DeepChainMap(*children))

    @staticmethod
    def mirror_attributes(children):
        return MirrorAttributes(**DeepChainMap(*children))

    @staticmethod
    def snapshot_attributes(children):
        return SnapshotAttributes(**DeepChainMap(*children))

    @staticmethod
    def target(children):
        return {"target": Target(**DeepChainMap(*children))}

    @staticmethod
    def targets(children):
        targets = []
        for child in children:
            targets.append(child["target"])
        return {"targets": targets}

    @staticmethod
    def device_metadata(children):
        return {"device_metadata": DeviceMetaData(**DeepChainMap(*children))}

    @staticmethod
    def load_event(children):
        return LoadEvent(**DeepChainMap(*children))

    def resume_event(self, children):
        data = DeepChainMap(*children)
        self._handle_no_data(data)
        return ResumeEvent(**data)

    def remove_event(self, children):
        data = DeepChainMap(*children)
        self._handle_no_data(data)
        return RemoveEvent(**data)

    @staticmethod
    def rename_event(children):
        return RenameEvent(**DeepChainMap(*children))

    def clear_event(self, children):
        data = DeepChainMap(*children)
        self._handle_no_data(data)
        return ClearEvent(**data)

    @staticmethod
    def update_event(children):
        return UpdateEvent(**DeepChainMap(*children))

    @staticmethod
    def remove_optional(children):
        return DeepChainMap(*children)

    @staticmethod
    def resume_optional(children):
        return DeepChainMap(*children)

    @staticmethod
    def clear_optional(children):
        return DeepChainMap(*children)

    @staticmethod
    def no_data(_):
        return {"no_data": True}

    @staticmethod
    @v_args(inline=True)
    def device_active_metadata(device_metadata):
        return {"device_active_metadata": device_metadata["device_metadata"]}

    @staticmethod
    @v_args(inline=True)
    def device_inactive_metadata(device_metadata):
        return {"device_inactive_metadata": device_metadata["device_metadata"]}

    @staticmethod
    def _handle_no_data(data):
        """
        If a no_data event happens (no information of the device is available) the event is still measured,
        but only with the devices name and uuid. This handles that special case.
        """
        if "no_data" in data:
            data["device_metadata"] = DeviceMetaDataMinimal(name=data["name"], uuid=data["uuid"])
            del data["name"]
            del data["uuid"]

    # Because mirror targets contains multiple devices it needs special parsing for that
    @staticmethod
    @v_args(inline=True)
    def mirror_mirror_device_name(_, name):
        return {"mirror_device_name": name}

    @staticmethod
    @v_args(inline=True)
    def mirror_mirror_device_status(_, status):
        return {"mirror_device_status": status}

    @staticmethod
    def mirror_mirror_device_row(children):
        return {"mirror_device_row": MirrorDevice(**DeepChainMap(*children))}

    @staticmethod
    def mirror_mirror_device_data(children):
        devices = []
        for child in children:
            # We might get mirror_device_data because of the manual left recursion
            if "mirror_device_row" in child:
                devices.append(child["mirror_device_row"])
            if "mirror_device_data" in child:
                devices.append(child["mirror_device_data"])
        return {"mirror_device_data": devices}


parser = Lark(DM_GRAMMAR)
transformer = DeviceMapperTransformer()

# The parser_fast only works on newer lark versions
parser_fast = Lark(DM_GRAMMAR, parser="lalr", transformer=transformer)


def parse(data: str, event: str):
    if version.Version(lark.__version__) >= version.Version("1.0.0"):
        out = parser_fast.parse(event + data)
    else:
        out = transformer.transform(parser.parse(event + data))
    # TODO: Provide sanity checks for object construction e.g a linear target should always have LinearAttributes
    return out



"""add_mb_refstate_column

Revision ID: ae898986c6e9
Revises: cc2630851a1f
Create Date: 2021-03-03 11:49:00.860132

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "ae898986c6e9"
down_revision = "cc2630851a1f"
branch_labels = None
depends_on = None


def upgrade(engine_name):
    globals()[f"upgrade_{engine_name}"]()


def downgrade(engine_name):
    globals()[f"downgrade_{engine_name}"]()


def upgrade_registrar():
    pass


def downgrade_registrar():
    pass


def upgrade_cloud_verifier():
    op.add_column("verifiermain", sa.Column("mb_refstate", sa.String(1000)))


def downgrade_cloud_verifier():
    op.drop_column("verifiermain", "mb_refstate")



# This package implements event log checking.
# This checking is factored into two stages:
# 1. a policy maps a convenient expression of intended state into
#    a test to apply to event logs and PCR contents;
# 2. that test is applied to various log&PCRs pairs.

# Basic policies are built in.
# Additional policies are dynamically loaded according to config.
from . import example



import json as json_module
from typing import IO, Any, Dict, List, Union

JSONType = Union[str, int, float, bool, None, Dict[str, Any], List[Any]]

_list_types = [list, tuple]
try:
    from sqlalchemy.engine.row import Row  # sqlalchemy >= 1.4.

    _list_types.append(Row)
except ModuleNotFoundError:
    try:
        from sqlalchemy.engine import RowProxy

        _list_types.append(RowProxy)
    except ModuleNotFoundError:
        pass


def bytes_to_str(data: Any) -> Any:
    if isinstance(data, (bytes, bytearray)):
        data = data.decode("utf-8")
    elif isinstance(data, dict):
        for _k, _v in data.items():
            data[_k] = bytes_to_str(_v)
    elif isinstance(data, tuple(_list_types)):
        _l = list(data)
        for _k, _v in enumerate(_l):
            _l[_k] = bytes_to_str(_v)
        data = _l

    return data


def dumps(obj: JSONType, **kwargs: Any) -> str:
    try:
        ret = json_module.dumps(obj, **kwargs)
    except TypeError:
        # dumps() from the built-it json module does not work with bytes,
        # so let's convert those to str if we get a TypeError exception.
        ret = json_module.dumps(bytes_to_str(obj), **kwargs)
    return ret


def dump(obj: JSONType, fp: IO[str], **kwargs: Any) -> None:
    try:
        json_module.dump(obj, fp, **kwargs)
    except TypeError:
        # dump() from the built-it json module does not work with bytes,
        # so let's convert those to str if we get a TypeError exception.
        json_module.dump(bytes_to_str(obj), fp, **kwargs)


def load(fp: Any, **kwargs: Any) -> Any:
    return json_module.load(fp, **kwargs)


def loads(s: Union[str, bytes], **kwargs: Any) -> Any:
    return json_module.loads(s, **kwargs)


# JSON pickler that fulfills SQLAlchemy requirements, from
# social-storage-sqlalchemy.
# https://github.com/python-social-auth/social-storage-sqlalchemy/commit/39d129
class JSONPickler:
    """JSON pickler wrapper around json lib since SQLAlchemy invokes
    dumps with extra positional parameters"""

    @classmethod
    def dumps(cls, value: JSONType, *_args: Any, **_kwargs: Any) -> str:
        # pylint: disable=unused-argument
        """Dumps the python value into a JSON string"""
        return dumps(value)

    @classmethod
    def loads(cls, value: Union[str, bytes]) -> Any:
        """Parses the JSON string and returns the corresponding python value"""
        return loads(value)



import re
import typing

from . import policies, tests

# The Example class is an example of a policy for checking boot event logs
# against reference state.  This policy checks:
# - that SecureBoot was enabled
# - that a good BIOS, shim, grub, and kernel were run
# - that only good keys are allowed
# - that all known bad keys are forbidden
# - the initial ramdisk contents and kernel command line were good
#
# This policy expects the reference state to be a dict created by `json.load`
# containing the following.
# scrtm_and_bios - list of allowed {
#    scrtm - digest for PCR 0 event type EV_S_CRTM_VERSION
#    platform_firmware - sequence of digest for PCR 0 event type EV_EFI_PLATFORM_FIRMWARE_BLOB
# }
# pk - list of allowed PK keys
# kek - list of allowed KEK keys
# db - list of allowed db keys
# dbx - list of required dbx keys
# mokdig - list of allowed digests of MoKList (PCR 14 EV_IPL)
# mokxdig - list of allowed digests of MoKListX (PCR 14 EV_IPL)
# kernels - list of allowed {
#   shim_authcode_sha256: 0xhex (for PCR 4 EV_EFI_BOOT_SERVICES_APPLICATION),
#   grub_authcode_sha256: 0xhex (for PCR 4 EV_EFI_BOOT_SERVICES_APPLICATION),
#   kernel_authcode_sha256: 0xhex (for PCR 4 EV_EFI_BOOT_SERVICES_APPLICATION),
#   initrd_plain_sha256: 0xhex (for PCR 9 EV_IPL),
#   kernel_cmdline: regex (for PCR 8 EV_IPL event.Event)
# }
# Here 0xhex is a string starting with '0x' and continuing with lowercase
# hex digits; in the case of a hash value, this includes leading zeros as
# needed to express the full number of bytes the hash function is defined
# to produce.
# A digest is a map from hash-algorithm-name (sha1 or sha256) to 0xhex.
# A key is {SignatureOwner: UUID, SignatureData: 0xhex}.

# First, define some helper functions for checking that the refstate is valid.
# They raise Exception when something invalid is encountered.

hex_pat = re.compile("0x[0-9a-f]+")


def hex_test(dat: typing.Any) -> bool:
    if isinstance(dat, str) and hex_pat.fullmatch(dat):
        return True
    raise Exception(f"{dat!r} is not 0x followed by some lowercase hex digits")


digest_type_test = tests.dict_test(tests.type_test(str), hex_test)

allowed_scrtm_and_bios_test = tests.obj_test(
    scrtm=digest_type_test, platform_firmware=tests.list_test(digest_type_test)
)

allowed_scrtm_and_bios_list_test = tests.list_test(allowed_scrtm_and_bios_test)

allowed_kernel_test = tests.obj_test(
    shim_authcode_sha256=hex_test,
    grub_authcode_sha256=hex_test,
    kernel_authcode_sha256=hex_test,
    initrd_plain_sha256=hex_test,
    kernel_cmdline=tests.type_test(str),
)

allowed_kernel_list_test = tests.list_test(allowed_kernel_test)


class Example(policies.Policy):
    relevant_pcr_indices = frozenset(list(range(10)) + [14])

    def get_relevant_pcrs(self) -> typing.FrozenSet[int]:
        return self.relevant_pcr_indices

    def refstate_to_test(self, refstate: policies.RefState) -> tests.Test:
        """Return the boot event log test corresponding to the given refstate
        The given refstate is expected to be Python data coming from `json.load`"""
        if not isinstance(refstate, dict):
            raise Exception(f"Expected refstate to be a Python dict but instead got this Python value: {refstate!r}")

        kernels = refstate.get("kernels")
        if not isinstance(kernels, list):
            raise Exception(f"refstate['kernels'] is {kernels!r} instead of a list")
        allowed_kernel_list_test(kernels)

        scrtm_and_bios_spec = refstate.get("scrtm_and_bios")
        if scrtm_and_bios_spec is None:
            raise Exception("refstate['scrtm_and_bios'] is missing")
        allowed_scrtm_and_bios_list_test(scrtm_and_bios_spec)
        scrtm_and_bios_test = tests.Or(
            *[
                tests.And(
                    tests.FieldTest("s_crtms", tests.TupleTest(tests.DigestTest(digest_strip0x(allowed["scrtm"])))),
                    tests.FieldTest(
                        "platform_firmware_blobs",
                        tests.TupleTest(*[tests.DigestTest(digest_strip0x(pf)) for pf in allowed["platform_firmware"]]),
                    ),
                )
                for allowed in scrtm_and_bios_spec
            ]
        )

        for req in ("pk", "kek", "db", "dbx", "mokdig", "mokxdig"):
            if req not in refstate:
                raise Exception(f"refstate lacks {req}")

        dispatcher = tests.Dispatcher(("PCRIndex", "EventType"))
        vd_driver_config = tests.VariableDispatch()
        vd_authority = tests.VariableDispatch()

        def bsa_test(kernel: dict) -> tests.Test:
            tt = [
                tests.DigestTest({"sha256": string_strip0x(kernel["shim_authcode_sha256"])}),
                tests.DigestTest({"sha256": string_strip0x(kernel["grub_authcode_sha256"])}),
                tests.DigestTest({"sha256": string_strip0x(kernel["kernel_authcode_sha256"])}),
            ]

            # In some scenarios the kernel gets measured twice
            tt2 = [
                tests.DigestTest({"sha256": string_strip0x(kernel["shim_authcode_sha256"])}),
                tests.DigestTest({"sha256": string_strip0x(kernel["grub_authcode_sha256"])}),
                tests.DigestTest({"sha256": string_strip0x(kernel["kernel_authcode_sha256"])}),
                tests.DigestTest({"sha256": string_strip0x(kernel["kernel_authcode_sha256"])}),
            ]
            return tests.Or(tests.TupleTest(*tt), tests.TupleTest(*tt2))

        events_final = tests.DelayToFields(
            tests.And(
                tests.Or(
                    *[
                        tests.FieldsTest(
                            bsas=bsa_test(kernel),
                            ipl9s=tests.TupleTest(
                                tests.DigestTest({"sha256": string_strip0x(kernel["initrd_plain_sha256"])})
                            ),
                            kernel_cmdlines=tests.TupleTest(
                                tests.RegExp("kernel_cmdline: " + kernel["kernel_cmdline"])
                            ),
                        )
                        for kernel in kernels
                    ]
                ),
                scrtm_and_bios_test,
            ),
            "kernel_cmdlines",
            "bsas",
            "ipl9s",
            "s_crtms",
            "platform_firmware_blobs",
        )
        # We only expect one EV_NO_ACTION event at the start.
        dispatcher.set((0, "EV_NO_ACTION"), tests.OnceTest(tests.AcceptAll()))
        dispatcher.set((0, "EV_S_CRTM_VERSION"), events_final.get("s_crtms"))
        dispatcher.set((0, "EV_EFI_PLATFORM_FIRMWARE_BLOB"), events_final.get("platform_firmware_blobs"))
        dispatcher.set((7, "EV_EFI_VARIABLE_DRIVER_CONFIG"), vd_driver_config)
        # tpm2-tools versions < 5.2 parsed the GUIDs wrong therefore we include a check for both here
        # For more information see: https://github.com/keylime/keylime/issues/1003
        secure_boot_test = tests.FieldTest("Enabled", tests.StringEqual("Yes"))
        vd_driver_config.set("61dfe48b-ca93-d211-aa0d-00e098032b8c", "SecureBoot", secure_boot_test)
        vd_driver_config.set("8be4df61-93ca-11d2-aa0d-00e098032b8c", "SecureBoot", secure_boot_test)
        pk_test = tests.OnceTest(
            tests.Or(
                tests.KeySubset("a159c0a5-e494-a74a-87b5-ab155c2bf072", sigs_strip0x(refstate["pk"])),
                tests.KeySubset("a5c059a1-94e4-4aa7-87b5-ab155c2bf072", sigs_strip0x(refstate["pk"])),
            )
        )
        vd_driver_config.set(
            "61dfe48b-ca93-d211-aa0d-00e098032b8c",
            "PK",
            pk_test,
        )
        vd_driver_config.set(
            "8be4df61-93ca-11d2-aa0d-00e098032b8c",
            "PK",
            pk_test,
        )

        kek_test = tests.OnceTest(
            tests.Or(
                tests.KeySubset("a159c0a5-e494-a74a-87b5-ab155c2bf072", sigs_strip0x(refstate["kek"])),
                tests.KeySubset("a5c059a1-94e4-4aa7-87b5-ab155c2bf072", sigs_strip0x(refstate["kek"])),
            )
        )
        vd_driver_config.set(
            "61dfe48b-ca93-d211-aa0d-00e098032b8c",
            "KEK",
            kek_test,
        )
        vd_driver_config.set(
            "8be4df61-93ca-11d2-aa0d-00e098032b8c",
            "KEK",
            kek_test,
        )
        db_test = tests.OnceTest(
            tests.Or(
                tests.KeySubset("a159c0a5-e494-a74a-87b5-ab155c2bf072", sigs_strip0x(refstate["db"])),
                tests.KeySubset("a5c059a1-94e4-4aa7-87b5-ab155c2bf072", sigs_strip0x(refstate["db"])),
            )
        )
        vd_driver_config.set(
            "cbb219d7-3a3d-9645-a3bc-dad00e67656f",
            "db",
            db_test,
        )
        vd_driver_config.set(
            "d719b2cb-3d3a-4596-a3bc-dad00e67656f",
            "db",
            db_test,
        )
        dbx_test = tests.OnceTest(
            tests.Or(
                tests.KeySuperset("2616c4c1-4c50-9240-aca9-41f936934328", sigs_strip0x(refstate["dbx"])),
                tests.KeySuperset("c1c41626-504c-4092-aca9-41f936934328", sigs_strip0x(refstate["dbx"])),
            )
        )
        vd_driver_config.set(
            "cbb219d7-3a3d-9645-a3bc-dad00e67656f",
            "dbx",
            dbx_test,
        )
        vd_driver_config.set(
            "d719b2cb-3d3a-4596-a3bc-dad00e67656f",
            "dbx",
            dbx_test,
        )
        dispatcher.set((7, "EV_EFI_VARIABLE_AUTHORITY"), vd_authority)
        # Assume that the cert that was used to verify the Shim is always trusted.
        # TODO: can we use the db entry for that instead of AcceptAll?
        vd_db_test = tests.OnceTest(tests.AcceptAll())
        vd_authority.set("cbb219d7-3a3d-9645-a3bc-dad00e67656f", "db", vd_db_test)
        vd_authority.set("d719b2cb-3d3a-4596-a3bc-dad00e67656f", "db", vd_db_test)
        # Accept all SbatLevels of the Shim, because we already checked the hash of the Shim itself.
        vd_sbat_level_test = tests.OnceTest(tests.AcceptAll())
        vd_authority.set("50ab5d60-46e0-0043-abb6-3dd810dd8b23", "SbatLevel", vd_sbat_level_test)
        vd_authority.set("605dab50-e046-4300-abb6-3dd810dd8b23", "SbatLevel", vd_sbat_level_test)
        # Accept all certificates that are used by the Shim to verify the next component,
        # because we already checked the hash of the Shim itself.
        vd_shim_test = tests.OnceTest(tests.AcceptAll())
        vd_authority.set("50ab5d60-46e0-0043-abb6-3dd810dd8b23", "Shim", vd_shim_test)
        vd_authority.set("605dab50-e046-4300-abb6-3dd810dd8b23", "Shim", vd_shim_test)

        # A list of allowed digests for firmware from device driver appears
        # in PCR2, event type EV_EFI_BOOT_SERVICES_DRIVER. Here we will just
        # accept everything.
        # This is fine because we do not use any other entry type from PCR 2 for validation.
        dispatcher.set((2, "EV_EFI_BOOT_SERVICES_DRIVER"), tests.AcceptAll())
        dispatcher.set(
            (1, "EV_EFI_VARIABLE_BOOT"),
            tests.Or(
                tests.VariableTest(
                    "61dfe48b-ca93-d211-aa0d-00e098032b8c", re.compile("BootOrder|Boot[0-9a-fA-F]+"), tests.AcceptAll()
                ),
                tests.VariableTest(
                    "8be4df61-93ca-11d2-aa0d-00e098032b8c", re.compile("BootOrder|Boot[0-9a-fA-F]+"), tests.AcceptAll()
                ),
            ),
        )
        dispatcher.set((4, "EV_EFI_ACTION"), tests.EvEfiActionTest(4))
        for pcr in range(8):
            dispatcher.set((pcr, "EV_SEPARATOR"), tests.EvSeperatorTest())

        dispatcher.set((4, "EV_EFI_BOOT_SERVICES_APPLICATION"), events_final.get("bsas"))
        dispatcher.set(
            (14, "EV_IPL"),
            tests.Or(
                tests.And(
                    tests.FieldTest("Event", tests.FieldTest("String", tests.StringEqual("MokList"))),
                    tests.DigestsTest(digests_strip0x(refstate["mokdig"])),
                ),
                tests.And(
                    tests.FieldTest("Event", tests.FieldTest("String", tests.StringEqual("MokListX"))),
                    tests.DigestsTest(digests_strip0x(refstate["mokxdig"])),
                ),
            ),
        )
        dispatcher.set(
            (9, "EV_IPL"),
            tests.Or(
                tests.FieldTest(
                    "Event", tests.FieldTest("String", tests.RegExp(r".*/loader/entries.*"))
                ),  # Ignore  Boot Loader Spec files
                tests.FieldTest("Event", tests.FieldTest("String", tests.RegExp(r".*/grub.*"))),
                tests.FieldTest("Event", tests.FieldTest("String", tests.RegExp(r".*/vmlinuz.*"))),
                tests.And(
                    tests.FieldTest(
                        "Event",
                        tests.FieldTest(
                            "String",
                            tests.Or(
                                tests.RegExp(r".*/initrd.*"),
                                tests.RegExp(r".*/initramfs.*"),
                            ),
                        ),
                    ),
                    events_final.get("ipl9s"),
                ),
            ),
        )
        dispatcher.set(
            (8, "EV_IPL"),
            tests.FieldTest(
                "Event",
                tests.FieldTest(
                    "String",
                    tests.Or(
                        tests.RegExp("grub_cmd: .*", re.DOTALL),
                        tests.And(tests.RegExp("kernel_cmdline: .*"), events_final.get("kernel_cmdlines")),
                    ),
                ),
            ),
        )
        dispatcher.set((5, "EV_EFI_ACTION"), tests.EvEfiActionTest(5))
        # Accept all UEFI_GPT_DATA. We only expect one entry for that.
        dispatcher.set((5, "EV_EFI_GPT_EVENT"), tests.OnceTest(tests.AcceptAll()))
        events_test = tests.FieldTest(
            "events",
            tests.And(events_final.get_initializer(), tests.IterateTest(dispatcher, show_elt=True), events_final),
            show_name=False,
        )
        return events_test


def string_strip0x(con: str) -> str:
    if con.startswith("0x"):
        return con[2:]
    raise Exception(f"{con!r} does not start with 0x")


def digest_strip0x(digest: typing.Dict[str, str]) -> tests.Digest:
    digest_type_test(digest)
    return {alg: string_strip0x(val) for alg, val in digest.items()}


def digests_strip0x(digests: typing.List[typing.Dict[str, str]]) -> typing.List[tests.Digest]:
    tests.type_test(list)(digests)
    return map(digest_strip0x, digests)


def sig_strip0x(sig: typing.Dict[str, str]) -> tests.Signature:
    tests.obj_test(SignatureOwner=tests.type_test(str), SignatureData=tests.type_test(str))(sig)
    return dict(SignatureOwner=sig["SignatureOwner"], SignatureData=string_strip0x(sig["SignatureData"]))


def sigs_strip0x(sigs: typing.Iterable[typing.Dict[str, str]]) -> typing.List[tests.Signature]:
    tests.type_test(typing.Iterable)(sigs)
    return map(sig_strip0x, sigs)


policies.register("example", Example())



import tempfile
import unittest
from unittest.mock import mock_open, patch

from keylime import secure_mount


class TestSecureMount(unittest.TestCase):
    def setUp(self):
        """Remove global state from secure_mount module."""
        # pylint: disable=protected-access
        secure_mount._MOUNTED = []

    @patch("keylime.secure_mount.logger")
    def test_check_mounted_not_found(self, _logger_mock):
        """Test when secdir is not mounted."""
        read_data = "23 106 0:21 / /proc rw,nosuid,nodev,noexec,relatime shared:26 - proc proc rw"
        with patch("keylime.secure_mount.open", mock_open(read_data=read_data)):
            self.assertFalse(secure_mount.check_mounted("/secdir"))

    @patch("keylime.secure_mount.logger")
    def test_check_mounted_found(self, _logger_mock):
        """Test when secdir is mounted."""
        read_data = (
            "23 106 0:21 / /proc rw,nosuid,nodev,noexec,relatime shared:26 - proc proc rw\n"
            "303 154 0:69 / /secdir rw,relatime shared:130 - tmpfs tmpfs rw,size=1024k,mode=700,inode64"
        )
        with patch("keylime.secure_mount.open", mock_open(read_data=read_data)):
            self.assertTrue(secure_mount.check_mounted("/secdir"))

    @patch("keylime.secure_mount.logger")
    def test_check_mounted_found_zero_optional_fields(self, _logger_mock):
        """Test when secdir is mounted when there are no optional fields."""
        read_data = (
            "23 106 0:21 / /proc rw,nosuid,nodev,noexec,relatime shared:26 - proc proc rw\n"
            "303 154 0:69 / /secdir rw,relatime - tmpfs tmpfs rw,size=1024k,mode=700,inode64"
        )
        with patch("keylime.secure_mount.open", mock_open(read_data=read_data)):
            self.assertTrue(secure_mount.check_mounted("/secdir"))

    @patch("keylime.secure_mount.logger")
    def test_check_mounted_found_extra_optional_fields(self, _logger_mock):
        """Test when secdir is mounted when there are extra optional fields."""
        read_data = (
            "23 106 0:21 / /proc rw,nosuid,nodev,noexec,relatime shared:26 - proc proc rw\n"
            "303 154 0:69 / /secdir rw,relatime shared:130 extra:1 - tmpfs tmpfs rw,size=1024k,mode=700,inode64"
        )
        with patch("keylime.secure_mount.open", mock_open(read_data=read_data)):
            self.assertTrue(secure_mount.check_mounted("/secdir"))

    @patch("keylime.secure_mount.logger")
    def test_check_mounted_found_wrong_fs(self, _logger_mock):
        """Test when secdir is mounted but under a wrong fs."""
        read_data = (
            "23 106 0:21 / /proc rw,nosuid,nodev,noexec,relatime shared:26 - proc proc rw\n"
            "303 154 0:69 / /secdir rw,relatime shared:130 - btrfs /dev/sda2 rw"
        )
        with patch("keylime.secure_mount.open", mock_open(read_data=read_data)):
            with self.assertRaises(Exception) as e:
                secure_mount.check_mounted("/secdir")
            self.assertTrue("wrong file system" in str(e.exception))

    @patch("keylime.secure_mount.logger")
    def test_check_mounted_found_spaces(self, _logger_mock):
        """Test when secdir is mounted and contains spaces."""
        read_data = (
            "23 106 0:21 / /proc rw,nosuid,nodev,noexec,relatime shared:26 - proc proc rw\n"
            r"303 154 0:69 / /sec\040dir rw,relatime shared:130 - tmpfs tmpfs rw,size=1024k,mode=700,inode64"
        )
        with patch("keylime.secure_mount.open", mock_open(read_data=read_data)):
            self.assertTrue(secure_mount.check_mounted("/sec dir"))

    @patch("keylime.secure_mount.logger")
    def test_check_mounted_wrong_format(self, _logger_mock):
        """Test when the mount info lines are wrong."""
        read_data = "invalid line"
        with patch("keylime.secure_mount.open", mock_open(read_data=read_data)):
            with self.assertRaises(Exception) as e:
                secure_mount.check_mounted("/secdir")
            self.assertTrue("cannot be parsed" in str(e.exception))

    @patch("keylime.secure_mount.logger")
    @patch("keylime.secure_mount.config")
    def test_check_mount_no_secure(self, config_mock, _logger_mock):
        """Test when mounted outside tmpfs."""
        config_mock.MOUNT_SECURE = False

        with tempfile.TemporaryDirectory() as tmpdirname:
            config_mock.WORK_DIR = tmpdirname
            self.assertEqual(secure_mount.mount(), f"{tmpdirname}/tmpfs-dev")
            # pylint: disable=protected-access
            self.assertEqual(secure_mount._MOUNTED, [])

    @patch("keylime.secure_mount.check_mounted")
    @patch("keylime.secure_mount.config")
    def test_check_mount_secure_already_mounted(self, config_mock, check_mounted_mock):
        """Test when mounting in tmpfs but is already present."""
        config_mock.MOUNT_SECURE = True
        check_mounted_mock.return_value = True

        with tempfile.TemporaryDirectory() as tmpdirname:
            config_mock.WORK_DIR = tmpdirname
            self.assertEqual(secure_mount.mount(), f"{tmpdirname}/secure")
            # pylint: disable=protected-access
            self.assertEqual(secure_mount._MOUNTED, [])

    @patch("keylime.secure_mount.logger")
    @patch("keylime.secure_mount.check_mounted")
    @patch("keylime.secure_mount.config")
    @patch("keylime.secure_mount.os.path.exists")
    @patch("keylime.secure_mount.cmd_exec")
    def test_check_mount_secure_already_created(
        self, _cmd_exec_mock, exists_mock, config_mock, check_mounted_mock, _logger_mock
    ):
        """Test when mounting in tmpfs but the mount point is present."""
        exists_mock.return_value = True
        config_mock.MOUNT_SECURE = True
        check_mounted_mock.return_value = False

        with tempfile.TemporaryDirectory() as tmpdirname:
            config_mock.WORK_DIR = tmpdirname
            self.assertEqual(secure_mount.mount(), f"{tmpdirname}/secure")
            # pylint: disable=protected-access
            self.assertEqual(secure_mount._MOUNTED, [f"{tmpdirname}/secure"])

    @patch("keylime.secure_mount.logger")
    @patch("keylime.secure_mount.check_mounted")
    @patch("keylime.secure_mount.config")
    @patch("keylime.secure_mount.os.path.exists")
    @patch("keylime.secure_mount.os.makedirs")
    @patch("keylime.secure_mount.cmd_exec")
    def test_check_mount(
        self,
        _cmd_exec_mock,
        _makedirs_mock,
        exists_mock,
        config_mock,
        check_mounted_mock,
        _logger_mock,
    ):
        """Test when mounting in tmpfs but the mount point is not present."""
        exists_mock.return_value = False
        config_mock.MOUNT_SECURE = True
        check_mounted_mock.return_value = False

        with tempfile.TemporaryDirectory() as tmpdirname:
            config_mock.WORK_DIR = tmpdirname
            self.assertEqual(secure_mount.mount(), f"{tmpdirname}/secure")
            # pylint: disable=protected-access
            self.assertEqual(secure_mount._MOUNTED, [f"{tmpdirname}/secure"])

    def test_check_umount_empty(self):
        """Test umount when there are nothing to clean."""
        secure_mount.umount()
        # pylint: disable=protected-access
        self.assertEqual(secure_mount._MOUNTED, [])

    @patch("keylime.secure_mount.logger")
    @patch("keylime.secure_mount.check_mounted")
    @patch("keylime.secure_mount.cmd_exec")
    def test_check_umount_mount(self, cmd_exec_mock, check_mounted_mock, logger_mock):
        """Test umount for a single mount."""
        # pylint: disable=protected-access
        secure_mount._MOUNTED = ["/secdir"]
        check_mounted_mock.return_value = True
        cmd_exec_mock.run.return_value = {"code": 0}
        secure_mount.umount()
        self.assertEqual(secure_mount._MOUNTED, [])
        cmd_exec_mock.run.assert_called_once()
        logger_mock.error.assert_not_called()
        logger_mock.warning.assert_not_called()

    @patch("keylime.secure_mount.logger")
    @patch("keylime.secure_mount.check_mounted")
    @patch("keylime.secure_mount.cmd_exec")
    def test_check_umount_mount_busy(self, cmd_exec_mock, check_mounted_mock, logger_mock):
        """Test umount for a single mount that fail."""
        # pylint: disable=protected-access
        secure_mount._MOUNTED = ["/secdir"]
        check_mounted_mock.return_value = True
        cmd_exec_mock.run.return_value = {"code": 1, "reterr": "Device busy"}
        secure_mount.umount()
        self.assertEqual(secure_mount._MOUNTED, [])
        cmd_exec_mock.run.assert_called_once()
        logger_mock.error.assert_called_once()
        logger_mock.warning.assert_not_called()

    @patch("keylime.secure_mount.logger")
    @patch("keylime.secure_mount.check_mounted")
    @patch("keylime.secure_mount.cmd_exec")
    def test_check_umount_umount(self, cmd_exec_mock, check_mounted_mock, logger_mock):
        """Test umount for a single umount and no creation."""
        # pylint: disable=protected-access
        secure_mount._MOUNTED = ["/secdir"]
        check_mounted_mock.return_value = False
        secure_mount.umount()
        self.assertEqual(secure_mount._MOUNTED, [])
        cmd_exec_mock.run.assert_not_called()
        logger_mock.error.assert_not_called()
        logger_mock.warning.assert_called_once()


if __name__ == "__main__":
    unittest.main()



import base64
import unittest

from cryptography.hazmat.backends import default_backend
from cryptography.x509 import load_der_x509_certificate

from keylime.tpm.tpm2_objects import (
    OA_ADMINWITHPOLICY,
    OA_DECRYPT,
    OA_ENCRYPTEDDUPLICATION,
    OA_FIXEDPARENT,
    OA_FIXEDTPM,
    OA_NODA,
    OA_RESTRICTED,
    OA_SENSITIVEDATAORIGIN,
    OA_SIGN_ENCRYPT,
    OA_STCLEAR,
    OA_USERWITHAUTH,
    ek_low_tpm2b_public_from_pubkey,
    get_tpm2b_public_name,
    get_tpm2b_public_object_attributes,
    object_attributes_description,
    pubkey_from_tpm2b_public,
)


class TestTpm2Objects(unittest.TestCase):
    def test_get_tpm2b_public_name(self):
        test_pub = base64.b64decode(
            "ARgAAQALAAUAcgAAABAAFAALCAAAAAAAAQDJBIF+SxeEt8TAwcnMZIvJWs3luBARcI"
            "HXC7I/XH7ZXbwLyispm/tpvhRw0w60JbwF4om1LbApQbG9cWR7AOi3ykv5bOgszsIG"
            "DOYJNfWuylW2uQBvMPEeF+ysrCjFTl5HOhXEpaz+E//juoKS2Jh9zYr2kt8rnGAJyj"
            "a10LUsYNt4h6eyeLVrsZIckkKP4tZwPOokfdX+6YCtGy5Y1buTvBSGNWa+VGo6hZVD"
            "649mg6EHyv0geSHXojx0Iqjsl/NQXzOCvyuaf6CBu9pkiIZCePlrl2uD1tXEdX0ipB"
            "B9Fppc/5cJQ2NyJOuvi4MUK5y38QpwnZwd4Utr2WdyEPoF"
        )
        test_pub_correct_name = "000b347dbfebe5bdbc55f6782a3cba91610f9d1b554a1aef07b4db28cf36da939009"
        new_name = get_tpm2b_public_name(test_pub)
        self.assertEqual(new_name, test_pub_correct_name)

    def test_get_tpm2b_public_object_attributes(self):
        test_pub = base64.b64decode(
            "ARgAAQALAAUAcgAAABAAFAALCAAAAAAAAQDJBIF+SxeEt8TAwcnMZIvJWs3luBARcI"
            "HXC7I/XH7ZXbwLyispm/tpvhRw0w60JbwF4om1LbApQbG9cWR7AOi3ykv5bOgszsIG"
            "DOYJNfWuylW2uQBvMPEeF+ysrCjFTl5HOhXEpaz+E//juoKS2Jh9zYr2kt8rnGAJyj"
            "a10LUsYNt4h6eyeLVrsZIckkKP4tZwPOokfdX+6YCtGy5Y1buTvBSGNWa+VGo6hZVD"
            "649mg6EHyv0geSHXojx0Iqjsl/NQXzOCvyuaf6CBu9pkiIZCePlrl2uD1tXEdX0ipB"
            "B9Fppc/5cJQ2NyJOuvi4MUK5y38QpwnZwd4Utr2WdyEPoF"
        )
        expected_attributes = (
            OA_RESTRICTED | OA_USERWITHAUTH | OA_SIGN_ENCRYPT | OA_FIXEDTPM | OA_FIXEDPARENT | OA_SENSITIVEDATAORIGIN
        )
        new_attributes = get_tpm2b_public_object_attributes(test_pub)
        self.assertEqual(new_attributes, expected_attributes)

    # Testing tpm2b_public_from_pubkey
    # These example certificates were standard EK certificates from a valid TPM,
    #  so these fields are selected according to
    #  TCG EK Credential Profile For TPM Family 2.0
    #  Level 0, Version 2.3, Revision 2"
    # Both are from the Low Ranges
    # The RSA set is according to Template L-1, section B.3.3
    # The EC set is according to Template L-2, section B.3.4

    def test_tpm2b_public_from_pubkey_rsa(self):
        test_rsa_cert = base64.b64decode(
            "MIIEnDCCA4SgAwIBAgIEL8wtHjANBgkqhkiG9w0BAQsFADCBgzELMAkGA1UEBhMCRE"
            "UxITAfBgNVBAoMGEluZmluZW9uIFRlY2hub2xvZ2llcyBBRzEaMBgGA1UECwwRT1BU"
            "SUdBKFRNKSBUUE0yLjAxNTAzBgNVBAMMLEluZmluZW9uIE9QVElHQShUTSkgUlNBIE"
            "1hbnVmYWN0dXJpbmcgQ0EgMDM1MB4XDTE4MDMwMTE0MTkzM1oXDTMzMDMwMTE0MTkz"
            "M1owADCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBALaIriXJCSUKdvWRDY"
            "dRbtdTK8i7eCJwHV8NhQ8Cor8NKoVmrnOdDhGqXlrKyJTueA9D2P4yQlWZI+tD9PCV"
            "CHCQiGmqxxHQXgzCzx6z+57HTUNPDi16K6ZFPNs3UkhAQxeGLOy36XD35zpfgadtvc"
            "lxJC8L+UgKfXVAM3/oMj4cDXa4cbVKhlfIQXD9OhcNjvESPWVFw0dj7Q6HM0jEkezM"
            "ew5sJ3I+LET1cIIhUlXvX8fWLu2MHx9+6LIBjkN8SuMLjKBQZjh+rEbHoFuG7Ib9pN"
            "ucrPAycid4EBBQB65j9irZ8C+ZdUUkKM5hsDhcenm/0AdfqAGXsFtsEa8DuDECAwEA"
            "AaOCAZgwggGUMFsGCCsGAQUFBwEBBE8wTTBLBggrBgEFBQcwAoY/aHR0cDovL3BraS"
            "5pbmZpbmVvbi5jb20vT3B0aWdhUnNhTWZyQ0EwMzUvT3B0aWdhUnNhTWZyQ0EwMzUu"
            "Y3J0MA4GA1UdDwEB/wQEAwIAIDBYBgNVHREBAf8ETjBMpEowSDEWMBQGBWeBBQIBDA"
            "tpZDo0OTQ2NTgwMDEaMBgGBWeBBQICDA9TTEIgOTY3MCBUUE0yLjAxEjAQBgVngQUC"
            "AwwHaWQ6MDczZjAMBgNVHRMBAf8EAjAAMFAGA1UdHwRJMEcwRaBDoEGGP2h0dHA6Ly"
            "9wa2kuaW5maW5lb24uY29tL09wdGlnYVJzYU1mckNBMDM1L09wdGlnYVJzYU1mckNB"
            "MDM1LmNybDAVBgNVHSAEDjAMMAoGCCqCFABEARQBMB8GA1UdIwQYMBaAFM53FTtuEQ"
            "ykrilxoJhR70mTJiAqMBAGA1UdJQQJMAcGBWeBBQgBMCEGA1UdCQQaMBgwFgYFZ4EF"
            "AhAxDTALDAMyLjACAQACAXQwDQYJKoZIhvcNAQELBQADggEBAIJ7pvW3yj2wAHO1fq"
            "zOeKg/xQjBMZ2hdpqVmhc+gU7F7zCMF85iWodISkThp9aa6p7VptkNcp5BNE1ojx+3"
            "1aJZRAFTCV0b0QxKXELTVsQLvBVmKGtFuaP3FPDVJYIOnQtb8uF+2LduF5P9K6oXdF"
            "TFuh1kG8GU/UUnltA7h6u2qhnj5uvFEDz7pxX1lt/GbI1nTYB+0SYtveIglpFyZK71"
            "0FH9UAvvR8byEbK+adE+teBUOexdXhTC1ZmPZmTvHSqmeRV3UTZFZRnyOTBnN8QlN0"
            "pMVmwFTak931PqxV0xOSXkMcvTre39jzkhEJ+VMb5EOMFfsVn+b4snob9jank="
        )
        correct_rsa_obj = base64.b64decode(
            "AToAAQALAAMAsgAgg3GXZ0SEs/gakMyNRqXXJP1S124GUgtk8qHaGzMUaaoABgCAAE"
            "MAEAgAAAAAAAEAtoiuJckJJQp29ZENh1Fu11MryLt4InAdXw2FDwKivw0qhWauc50O"
            "EapeWsrIlO54D0PY/jJCVZkj60P08JUIcJCIaarHEdBeDMLPHrP7nsdNQ08OLXorpk"
            "U82zdSSEBDF4Ys7LfpcPfnOl+Bp229yXEkLwv5SAp9dUAzf+gyPhwNdrhxtUqGV8hB"
            "cP06Fw2O8RI9ZUXDR2PtDoczSMSR7Mx7Dmwncj4sRPVwgiFSVe9fx9Yu7YwfH37osg"
            "GOQ3xK4wuMoFBmOH6sRsegW4bshv2k25ys8DJyJ3gQEFAHrmP2KtnwL5l1RSQozmGw"
            "OFx6eb/QB1+oAZewW2wRrwO4MQ=="
        )
        test_rsa_cert = load_der_x509_certificate(test_rsa_cert, backend=default_backend())
        new_rsa_obj = ek_low_tpm2b_public_from_pubkey(test_rsa_cert.public_key())
        self.assertEqual(new_rsa_obj.hex(), correct_rsa_obj.hex())

    def test_tpm2b_public_from_pubkey_ec(self):
        test_ec_cert = base64.b64decode(
            "MIIDEDCCAragAwIBAgIEcYSJiTAKBggqhkjOPQQDAjCBgzELMAkGA1UEBhMCREUxIT"
            "AfBgNVBAoMGEluZmluZW9uIFRlY2hub2xvZ2llcyBBRzEaMBgGA1UECwwRT1BUSUdB"
            "KFRNKSBUUE0yLjAxNTAzBgNVBAMMLEluZmluZW9uIE9QVElHQShUTSkgRUNDIE1hbn"
            "VmYWN0dXJpbmcgQ0EgMDM1MB4XDTE4MDMwMTE0MTkxNloXDTMzMDMwMTE0MTkxNlow"
            "ADBZMBMGByqGSM49AgEGCCqGSM49AwEHA0IABNK9AtBnW5bwNG2ZIWDrM8w/h03Ht2"
            "lp3MUosV05DeBHWZEZfmKsHMBqpqDsIKkEgclQawA4BFR5YUvSdrSUDTGjggGYMIIB"
            "lDBbBggrBgEFBQcBAQRPME0wSwYIKwYBBQUHMAKGP2h0dHA6Ly9wa2kuaW5maW5lb2"
            "4uY29tL09wdGlnYUVjY01mckNBMDM1L09wdGlnYUVjY01mckNBMDM1LmNydDAOBgNV"
            "HQ8BAf8EBAMCAAgwWAYDVR0RAQH/BE4wTKRKMEgxFjAUBgVngQUCAQwLaWQ6NDk0Nj"
            "U4MDAxGjAYBgVngQUCAgwPU0xCIDk2NzAgVFBNMi4wMRIwEAYFZ4EFAgMMB2lkOjA3"
            "M2YwDAYDVR0TAQH/BAIwADBQBgNVHR8ESTBHMEWgQ6BBhj9odHRwOi8vcGtpLmluZm"
            "luZW9uLmNvbS9PcHRpZ2FFY2NNZnJDQTAzNS9PcHRpZ2FFY2NNZnJDQTAzNS5jcmww"
            "FQYDVR0gBA4wDDAKBggqghQARAEUATAfBgNVHSMEGDAWgBQ2WY8i7ITDxPZA0hwWfQ"
            "uRE3uQpDAQBgNVHSUECTAHBgVngQUIATAhBgNVHQkEGjAYMBYGBWeBBQIQMQ0wCwwD"
            "Mi4wAgEAAgF0MAoGCCqGSM49BAMCA0gAMEUCIQCdCv3+G+KsM4OiT3SgKqvE8r5ktD"
            "I5elC9xTmS9mDA3AIgcckalMvQVTst1pGMEyAI+OoXTnYA1sBRm27WJ6sZag8="
        )
        correct_ec_obj = base64.b64decode(
            "AHoAIwALAAMAsgAgg3GXZ0SEs/gakMyNRqXXJP1S124GUgtk8qHaGzMUaaoABgCAAE"
            "MAEAADABAAINK9AtBnW5bwNG2ZIWDrM8w/h03Ht2lp3MUosV05DeBHACBZkRl+Yqwc"
            "wGqmoOwgqQSByVBrADgEVHlhS9J2tJQNMQ=="
        )
        test_ec_cert = load_der_x509_certificate(test_ec_cert, backend=default_backend())
        new_ec_obj = ek_low_tpm2b_public_from_pubkey(test_ec_cert.public_key())
        self.assertEqual(new_ec_obj.hex(), correct_ec_obj.hex())

    def test_pubkey_from_tpm2b_public_rsa(self):
        test_rsa_cert = base64.b64decode(
            "MIIEnDCCA4SgAwIBAgIEL8wtHjANBgkqhkiG9w0BAQsFADCBgzELMAkGA1UEBhMCRE"
            "UxITAfBgNVBAoMGEluZmluZW9uIFRlY2hub2xvZ2llcyBBRzEaMBgGA1UECwwRT1BU"
            "SUdBKFRNKSBUUE0yLjAxNTAzBgNVBAMMLEluZmluZW9uIE9QVElHQShUTSkgUlNBIE"
            "1hbnVmYWN0dXJpbmcgQ0EgMDM1MB4XDTE4MDMwMTE0MTkzM1oXDTMzMDMwMTE0MTkz"
            "M1owADCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBALaIriXJCSUKdvWRDY"
            "dRbtdTK8i7eCJwHV8NhQ8Cor8NKoVmrnOdDhGqXlrKyJTueA9D2P4yQlWZI+tD9PCV"
            "CHCQiGmqxxHQXgzCzx6z+57HTUNPDi16K6ZFPNs3UkhAQxeGLOy36XD35zpfgadtvc"
            "lxJC8L+UgKfXVAM3/oMj4cDXa4cbVKhlfIQXD9OhcNjvESPWVFw0dj7Q6HM0jEkezM"
            "ew5sJ3I+LET1cIIhUlXvX8fWLu2MHx9+6LIBjkN8SuMLjKBQZjh+rEbHoFuG7Ib9pN"
            "ucrPAycid4EBBQB65j9irZ8C+ZdUUkKM5hsDhcenm/0AdfqAGXsFtsEa8DuDECAwEA"
            "AaOCAZgwggGUMFsGCCsGAQUFBwEBBE8wTTBLBggrBgEFBQcwAoY/aHR0cDovL3BraS"
            "5pbmZpbmVvbi5jb20vT3B0aWdhUnNhTWZyQ0EwMzUvT3B0aWdhUnNhTWZyQ0EwMzUu"
            "Y3J0MA4GA1UdDwEB/wQEAwIAIDBYBgNVHREBAf8ETjBMpEowSDEWMBQGBWeBBQIBDA"
            "tpZDo0OTQ2NTgwMDEaMBgGBWeBBQICDA9TTEIgOTY3MCBUUE0yLjAxEjAQBgVngQUC"
            "AwwHaWQ6MDczZjAMBgNVHRMBAf8EAjAAMFAGA1UdHwRJMEcwRaBDoEGGP2h0dHA6Ly"
            "9wa2kuaW5maW5lb24uY29tL09wdGlnYVJzYU1mckNBMDM1L09wdGlnYVJzYU1mckNB"
            "MDM1LmNybDAVBgNVHSAEDjAMMAoGCCqCFABEARQBMB8GA1UdIwQYMBaAFM53FTtuEQ"
            "ykrilxoJhR70mTJiAqMBAGA1UdJQQJMAcGBWeBBQgBMCEGA1UdCQQaMBgwFgYFZ4EF"
            "AhAxDTALDAMyLjACAQACAXQwDQYJKoZIhvcNAQELBQADggEBAIJ7pvW3yj2wAHO1fq"
            "zOeKg/xQjBMZ2hdpqVmhc+gU7F7zCMF85iWodISkThp9aa6p7VptkNcp5BNE1ojx+3"
            "1aJZRAFTCV0b0QxKXELTVsQLvBVmKGtFuaP3FPDVJYIOnQtb8uF+2LduF5P9K6oXdF"
            "TFuh1kG8GU/UUnltA7h6u2qhnj5uvFEDz7pxX1lt/GbI1nTYB+0SYtveIglpFyZK71"
            "0FH9UAvvR8byEbK+adE+teBUOexdXhTC1ZmPZmTvHSqmeRV3UTZFZRnyOTBnN8QlN0"
            "pMVmwFTak931PqxV0xOSXkMcvTre39jzkhEJ+VMb5EOMFfsVn+b4snob9jank="
        )
        test_rsa_cert = load_der_x509_certificate(test_rsa_cert, backend=default_backend())
        correct_rsa_obj = base64.b64decode(
            "AToAAQALAAMAsgAgg3GXZ0SEs/gakMyNRqXXJP1S124GUgtk8qHaGzMUaaoABgCAAE"
            "MAEAgAAAAAAAEAtoiuJckJJQp29ZENh1Fu11MryLt4InAdXw2FDwKivw0qhWauc50O"
            "EapeWsrIlO54D0PY/jJCVZkj60P08JUIcJCIaarHEdBeDMLPHrP7nsdNQ08OLXorpk"
            "U82zdSSEBDF4Ys7LfpcPfnOl+Bp229yXEkLwv5SAp9dUAzf+gyPhwNdrhxtUqGV8hB"
            "cP06Fw2O8RI9ZUXDR2PtDoczSMSR7Mx7Dmwncj4sRPVwgiFSVe9fx9Yu7YwfH37osg"
            "GOQ3xK4wuMoFBmOH6sRsegW4bshv2k25ys8DJyJ3gQEFAHrmP2KtnwL5l1RSQozmGw"
            "OFx6eb/QB1+oAZewW2wRrwO4MQ=="
        )
        new_rsa_pubkey = pubkey_from_tpm2b_public(correct_rsa_obj)
        correct_rsa_pubkey = test_rsa_cert.public_key()
        new_rsa_pubkey_n = new_rsa_pubkey.public_numbers()
        correct_rsa_pubkey_n = correct_rsa_pubkey.public_numbers()
        self.assertEqual(new_rsa_pubkey.key_size, correct_rsa_pubkey.key_size)
        self.assertEqual(new_rsa_pubkey_n.e, correct_rsa_pubkey_n.e)  # pylint: disable=no-member
        self.assertEqual(new_rsa_pubkey_n.n, correct_rsa_pubkey_n.n)  # pylint: disable=no-member

    def test_pubkey_from_tpm2b_public_rsa_without_encryption(self):
        new_rsa_pubkey = pubkey_from_tpm2b_public(
            bytes.fromhex(
                "01180001000b00050072000000100014000b0800000000000100cac43903c6"
                "16bba049ce413c961c901b56181392c7999e672e6e5ecdd7a625d4702c3d78"
                "deac81e1372b0ca1894ac0f16add636bb53d3d5b112d8f3b169ccadef6bac0"
                "d909067d1ff81dae34b26cd538a52fa20ee7bbf3b16214417d35bde80cbb0f"
                "1b3267fd6211ecfb652f771f7eaeff560b91ef2f374ab1d37bba5a7a1c7cd4"
                "4961cdd7351ee060947f43244f45fc42ea6a1ea783aaa18dc8cce90d9a97f8"
                "da09e72637a0167fdbf4cc0d09f2f752d864d45bd34ed387acc0bcddca26c6"
                "1ebe9056013a35cd1d8011336af93579afa424fe50fd7e2b03270518505710"
                "82fcae891e2897e3117fd28bd03d2d2ffdfcfa0ff95f76af9383e3c9e59fe4"
                "dde753"
            )
        )
        new_rsa_pubkey_n = new_rsa_pubkey.public_numbers()

        self.assertEqual(new_rsa_pubkey.key_size, 2048)
        self.assertEqual(new_rsa_pubkey_n.e, 65537)  # pylint: disable=no-member
        self.assertEqual(
            str(new_rsa_pubkey_n.n),  # pylint: disable=no-member
            "255968986296679270326283402717529063492526907681140893873754141432"
            "890531031973586937300971300465026177966018575012122367284728088154"
            "485873651193407172159946655006581809152369460009001515677703036255"
            "837234635576083087037905135410736640524495731191518154258439490758"
            "531740360767515943902821573272461751306668217217601399605319344343"
            "524504419559281243744525835687758392857402638332592577865592671234"
            "679107983328133731582503713366603521336278457142403979969779706740"
            "010077961630324526931687863526905140593203113247551679416434551326"
            "587069716966112452602019925398408142602185862884082705845069125895"
            "71106286823536420841299",
        )

    def test_pubkey_from_tpm2b_public_ec(self):
        test_ec_cert = base64.b64decode(
            "MIIDEDCCAragAwIBAgIEcYSJiTAKBggqhkjOPQQDAjCBgzELMAkGA1UEBhMCREUxIT"
            "AfBgNVBAoMGEluZmluZW9uIFRlY2hub2xvZ2llcyBBRzEaMBgGA1UECwwRT1BUSUdB"
            "KFRNKSBUUE0yLjAxNTAzBgNVBAMMLEluZmluZW9uIE9QVElHQShUTSkgRUNDIE1hbn"
            "VmYWN0dXJpbmcgQ0EgMDM1MB4XDTE4MDMwMTE0MTkxNloXDTMzMDMwMTE0MTkxNlow"
            "ADBZMBMGByqGSM49AgEGCCqGSM49AwEHA0IABNK9AtBnW5bwNG2ZIWDrM8w/h03Ht2"
            "lp3MUosV05DeBHWZEZfmKsHMBqpqDsIKkEgclQawA4BFR5YUvSdrSUDTGjggGYMIIB"
            "lDBbBggrBgEFBQcBAQRPME0wSwYIKwYBBQUHMAKGP2h0dHA6Ly9wa2kuaW5maW5lb2"
            "4uY29tL09wdGlnYUVjY01mckNBMDM1L09wdGlnYUVjY01mckNBMDM1LmNydDAOBgNV"
            "HQ8BAf8EBAMCAAgwWAYDVR0RAQH/BE4wTKRKMEgxFjAUBgVngQUCAQwLaWQ6NDk0Nj"
            "U4MDAxGjAYBgVngQUCAgwPU0xCIDk2NzAgVFBNMi4wMRIwEAYFZ4EFAgMMB2lkOjA3"
            "M2YwDAYDVR0TAQH/BAIwADBQBgNVHR8ESTBHMEWgQ6BBhj9odHRwOi8vcGtpLmluZm"
            "luZW9uLmNvbS9PcHRpZ2FFY2NNZnJDQTAzNS9PcHRpZ2FFY2NNZnJDQTAzNS5jcmww"
            "FQYDVR0gBA4wDDAKBggqghQARAEUATAfBgNVHSMEGDAWgBQ2WY8i7ITDxPZA0hwWfQ"
            "uRE3uQpDAQBgNVHSUECTAHBgVngQUIATAhBgNVHQkEGjAYMBYGBWeBBQIQMQ0wCwwD"
            "Mi4wAgEAAgF0MAoGCCqGSM49BAMCA0gAMEUCIQCdCv3+G+KsM4OiT3SgKqvE8r5ktD"
            "I5elC9xTmS9mDA3AIgcckalMvQVTst1pGMEyAI+OoXTnYA1sBRm27WJ6sZag8="
        )
        correct_ec_obj = base64.b64decode(
            "AHoAIwALAAMAsgAgg3GXZ0SEs/gakMyNRqXXJP1S124GUgtk8qHaGzMUaaoABgCAAE"
            "MAEAADABAAINK9AtBnW5bwNG2ZIWDrM8w/h03Ht2lp3MUosV05DeBHACBZkRl+Yqwc"
            "wGqmoOwgqQSByVBrADgEVHlhS9J2tJQNMQ=="
        )
        test_ec_cert = load_der_x509_certificate(test_ec_cert, backend=default_backend())
        new_ec_pubkey = pubkey_from_tpm2b_public(correct_ec_obj)
        correct_ec_pubkey = test_ec_cert.public_key()
        new_ec_pubkey_n = new_ec_pubkey.public_numbers()
        correct_ec_pubkey_n = correct_ec_pubkey.public_numbers()
        self.assertEqual(new_ec_pubkey_n.curve.name, correct_ec_pubkey_n.curve.name)
        self.assertEqual(new_ec_pubkey_n.x, correct_ec_pubkey_n.x)
        self.assertEqual(new_ec_pubkey_n.y, correct_ec_pubkey_n.y)

    def test_pubkey_from_tpm2b_public_ec_without_encryption(self):
        new_ec_pubkey = pubkey_from_tpm2b_public(
            bytes.fromhex(
                "00580023000b00050072000000100018000b000300100020c74568135840f4"
                "97ad575ebeabe6d01f3f098b5a768111ab423d5f26b259a4f000205ec0f586"
                "b53e348bc916b43a015e6ceefd947d685e59ff65357499f2c4788cba"
            )
        )
        new_ec_pubkey_n = new_ec_pubkey.public_numbers()

        self.assertEqual(new_ec_pubkey_n.curve.name, "secp256r1")
        self.assertEqual(
            str(new_ec_pubkey_n.x),
            "90132887618692975484254453731651094410483286444689191401164175504334705501424",
        )
        self.assertEqual(
            str(new_ec_pubkey_n.y),
            "42858336962839421935559570622369777529185491150475599613778789950332157332666",
        )

    def test_object_attributes_description(self):
        with self.subTest(attrs="sign-encrypt"):
            val = object_attributes_description((OA_SIGN_ENCRYPT))
            self.assertEqual(val, "sign-encrypt")

        with self.subTest(attrs="<empty>"):
            val = object_attributes_description((0))
            self.assertEqual(val, "")

        with self.subTest(attrs="<all>"):
            val = object_attributes_description(
                (
                    OA_FIXEDTPM
                    | OA_STCLEAR
                    | OA_FIXEDPARENT
                    | OA_SENSITIVEDATAORIGIN
                    | OA_USERWITHAUTH
                    | OA_ADMINWITHPOLICY
                    | OA_NODA
                    | OA_ENCRYPTEDDUPLICATION
                    | OA_RESTRICTED
                    | OA_DECRYPT
                    | OA_SIGN_ENCRYPT
                )
            )
            self.assertEqual(
                val,
                "fixed-tpm | st-clear | fixed-parent | sensitive-data-origin | "
                "user-with-auth | admin-with-policy | no-da | "
                "encrypted-duplication | restricted | decrypt | sign-encrypt",
            )


if __name__ == "__main__":
    unittest.main()



import hashlib
import struct
from typing import Tuple, Union

import cryptography.hazmat.primitives.asymmetric.ec as crypto_ec
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.ec import (
    EllipticCurve,
    EllipticCurvePublicKey,
    EllipticCurvePublicNumbers,
)
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey, RSAPublicNumbers

pubkey_type = Union[RSAPublicKey, EllipticCurvePublicKey]


def _pack_in_tpm2b(val):
    return struct.pack(">H", len(val)) + val


# Algorithm constants
TPM2_ALG_NULL = 0x0010

TPM_ECC_NIST_P192 = 0x0001
TPM_ECC_NIST_P224 = 0x0002
TPM_ECC_NIST_P256 = 0x0003
TPM_ECC_NIST_P384 = 0x0004
TPM_ECC_NIST_P521 = 0x0005

TPM_ALG_RSA = 0x0001
TPM_ALG_ECC = 0x0023

TPM_ALG_SHA1 = 0x0004
TPM_ALG_SHA256 = 0x000B
TPM_ALG_SHA384 = 0x000C
TPM_ALG_SHA512 = 0x000D

TPM_ALG_AES = 0x0006
TPM_ALG_CFB = 0x0043


# These are the object attribute values important for EK certs
OA_FIXEDTPM = 0x00000002
OA_STCLEAR = 0x00000004
OA_FIXEDPARENT = 0x00000010
OA_SENSITIVEDATAORIGIN = 0x00000020
OA_USERWITHAUTH = 0x00000040
OA_ADMINWITHPOLICY = 0x00000080
OA_NODA = 0x00000400
OA_ENCRYPTEDDUPLICATION = 0x00000800
OA_RESTRICTED = 0x00010000
OA_DECRYPT = 0x00020000
OA_SIGN_ENCRYPT = 0x00040000


# These are some common object attribute values
# Source for AK_EXPECTED_ATTRS: tpm2-tools, tpm2_createak.c, set_key_algorithm
AK_EXPECTED_ATTRS = (
    OA_RESTRICTED | OA_USERWITHAUTH | OA_SIGN_ENCRYPT | OA_FIXEDTPM | OA_FIXEDPARENT | OA_SENSITIVEDATAORIGIN
)


class NonAsymAlgSpecificParameters:
    sym_algorithm = None
    sym_keybits = None
    sym_mode = None
    sym_details = None
    scheme_scheme = None
    scheme_details = None

    def __init__(self, sym_algorithm, sym_keybits, sym_mode, sym_details, scheme_scheme, scheme_details):
        self.sym_algorithm = sym_algorithm
        self.sym_keybits = sym_keybits
        self.sym_mode = sym_mode
        self.sym_details = sym_details
        self.scheme_scheme = scheme_scheme
        self.scheme_details = scheme_details

    def to_bytes(self):
        sym = struct.pack(">HHH", self.sym_algorithm, self.sym_keybits, self.sym_mode)
        scheme = struct.pack(">H", self.scheme_scheme)
        return sym + scheme


# These values come from "TCG EK Credential Profile For TPM Family 2.0;
#  Level 0, Version 2.3, Revision 2"
EK_LOW_NAMEALG = TPM_ALG_SHA256
EK_HIGH_SHA256_NAMEALG = EK_LOW_NAMEALG
EK_LOW_ATTRIBUTES = (
    OA_FIXEDTPM
    |
    # ~OA_STCLEAR |
    OA_FIXEDPARENT
    | OA_SENSITIVEDATAORIGIN
    |
    # ~OA_USERWITHAUTH |
    OA_ADMINWITHPOLICY
    |
    # ~OA_NODA |
    # ~OA_ENCRYPTEDDUPLICATION |
    OA_RESTRICTED
    | OA_DECRYPT
    # ~OA_SIGN_ENCRYPT,
)
EK_HIGH_ATTRIBUTES = (
    OA_FIXEDTPM
    |
    # ~OA_STCLEAR |
    OA_FIXEDPARENT
    | OA_SENSITIVEDATAORIGIN
    | OA_USERWITHAUTH
    | OA_ADMINWITHPOLICY
    |
    # ~OA_NODA |
    # ~OA_ENCRYPTEDDUPLICATION |
    OA_RESTRICTED
    | OA_DECRYPT
    # ~OA_SIGN_ENCRYPT,
)
# TPM2_PolicySecret(TPM_RH_ENDORSEMENT)
EK_LOW_AUTH_POLICY = bytes(
    [
        0x83,
        0x71,
        0x97,
        0x67,
        0x44,
        0x84,
        0xB3,
        0xF8,
        0x1A,
        0x90,
        0xCC,
        0x8D,
        0x46,
        0xA5,
        0xD7,
        0x24,
        0xFD,
        0x52,
        0xD7,
        0x6E,
        0x06,
        0x52,
        0x0B,
        0x64,
        0xF2,
        0xA1,
        0xDA,
        0x1B,
        0x33,
        0x14,
        0x69,
        0xAA,
    ]
)
EK_HIGH_SHA256_AUTH_POLICY = bytes(
    [
        0xCA,
        0x3D,
        0x0A,
        0x99,
        0xA2,
        0xB9,
        0x39,
        0x06,
        0xF7,
        0xA3,
        0x34,
        0x24,
        0x14,
        0xEF,
        0xCF,
        0xB3,
        0xA3,
        0x85,
        0xD4,
        0x4C,
        0xD1,
        0xFD,
        0x45,
        0x90,
        0x89,
        0xD1,
        0x9B,
        0x50,
        0x71,
        0xC0,
        0xB7,
        0xA0,
    ]
)
EK_LOW_NON_ASYM_ALG_PARMS = NonAsymAlgSpecificParameters(
    TPM_ALG_AES,
    128,
    TPM_ALG_CFB,
    0x00,  # NULL
    TPM2_ALG_NULL,
    0x00,  # NULL
)
EK_HIGH_SHA256_NON_ASYM_ALG_PARMS = EK_LOW_NON_ASYM_ALG_PARMS


def _curve_id_from_name(name: str) -> int:
    if name == "secp192r1":
        return TPM_ECC_NIST_P192
    if name == "secp224r1":
        return TPM_ECC_NIST_P224
    if name == "secp256r1":
        return TPM_ECC_NIST_P256
    if name == "secp384r1":
        return TPM_ECC_NIST_P384
    if name == "secp521r1":
        return TPM_ECC_NIST_P521

    raise ValueError(f"Invalid curve name {name} requested")


def _curve_from_curve_id(cid: int) -> EllipticCurve:
    if cid == TPM_ECC_NIST_P192:
        return crypto_ec.SECP192R1()
    if cid == TPM_ECC_NIST_P224:
        return crypto_ec.SECP224R1()
    if cid == TPM_ECC_NIST_P256:
        return crypto_ec.SECP256R1()
    if cid == TPM_ECC_NIST_P384:
        return crypto_ec.SECP384R1()
    if cid == TPM_ECC_NIST_P521:
        return crypto_ec.SECP521R1()

    raise ValueError(f"Invalid curve id {cid} requested")


def _extract_tpm2b(vals: bytes) -> Tuple[bytes, bytes]:
    (length,) = struct.unpack(">H", vals[0:2])
    # Ignore the length itself when returning
    vals = vals[2:]
    # Return first the currect buffer, and then the rest
    return (vals[:length], vals[length:])


def pubkey_from_tpm2b_public(public: bytes) -> pubkey_type:
    (public, rest) = _extract_tpm2b(public)
    if len(rest) != 0:
        raise ValueError("More in tpm2b_public than tpmt_public")
    # Extract type, [nameAlg], and [objectAttributes] (we don't care about the
    #  latter two)
    (alg_type, _, _) = struct.unpack(">HHI", public[0:8])
    # Ignore the authPolicy
    (_, sym_parms) = _extract_tpm2b(public[8:])
    # Ignore the non-asym-alg parameters
    (sym_mode,) = struct.unpack(">H", sym_parms[0:2])
    # Ignore the sym_mode and keybits (4 bytes), possibly symmetric (2) and sign
    #  scheme (2)
    to_skip = 4 + 2  # sym_mode, keybits and sign scheme
    if sym_mode != TPM2_ALG_NULL:
        to_skip = to_skip + 2
    asym_parms = sym_parms[to_skip:]

    # Handle fields
    if alg_type == TPM_ALG_RSA:
        (keybits, exponent) = struct.unpack(">HI", asym_parms[0:6])
        if exponent == 0:
            exponent = 65537
        (modulus, _) = _extract_tpm2b(asym_parms[6:])
        if (len(modulus) * 8) != keybits:
            raise ValueError(f"Misparsed either modulus or keybits: {len(modulus)}*8 != {keybits}")
        bmodulus = int.from_bytes(modulus, byteorder="big")

        numbers = RSAPublicNumbers(exponent, bmodulus)
        return numbers.public_key(backend=default_backend())

    if alg_type == TPM_ALG_ECC:
        (curve, _) = struct.unpack(">HH", asym_parms[0:4])
        asym_x = asym_parms[4:]
        curve = _curve_from_curve_id(curve)

        (x, asym_y) = _extract_tpm2b(asym_x)
        (y, rest) = _extract_tpm2b(asym_y)
        if len(rest) != 0:
            raise ValueError("Misparsed: more contents after X and Y")

        if (len(x) * 8) != curve.key_size:
            raise ValueError(f"Misparsed either X or curve: {len(x)}*8 != {curve.key_size}")
        if (len(y) * 8) != curve.key_size:
            raise ValueError(f"Misparsed either Y or curve curve: {len(y)}*8 != {curve.key_size}")

        bx = int.from_bytes(x, byteorder="big")
        by = int.from_bytes(y, byteorder="big")

        numbers = EllipticCurvePublicNumbers(bx, by, curve)
        return numbers.public_key(backend=default_backend())

    raise ValueError(f"Invalid tpm2b_public type: {alg_type}")


def tpm2b_public_from_pubkey(
    pubkey: pubkey_type, name_alg: int, attributes: int, auth_policy: bytes, parms: NonAsymAlgSpecificParameters
) -> bytes:
    """
    Returns a reconstructed TPM2B_PUBLIC from a public key.
    """
    if isinstance(pubkey, RSAPublicKey):
        alg_type = TPM_ALG_RSA

        numbers = pubkey.public_numbers()
        n = numbers.n.to_bytes((numbers.n.bit_length() + 7) // 8, byteorder="big")

        pub_e = numbers.e
        if pub_e == 65537:
            pub_e = 0

        algo_parms = struct.pack(">HI", pubkey.key_size, pub_e)
        unique = _pack_in_tpm2b(n)
    elif isinstance(pubkey, EllipticCurvePublicKey):
        alg_type = TPM_ALG_ECC

        numbers = pubkey.public_numbers()

        algo_parms = struct.pack(
            ">HH",
            _curve_id_from_name(numbers.curve.name),
            TPM2_ALG_NULL,
        )
        unique_x = numbers.x.to_bytes((numbers.x.bit_length() + 7) // 8, byteorder="big")
        unique_y = numbers.y.to_bytes((numbers.y.bit_length() + 7) // 8, byteorder="big")
        unique = _pack_in_tpm2b(unique_x) + _pack_in_tpm2b(unique_y)
    else:
        raise ValueError("Unsupported public key type")

    auth_policy = _pack_in_tpm2b(auth_policy)
    parameters = parms.to_bytes() + algo_parms

    tpmt_pub = (
        struct.pack(
            ">HHI",
            alg_type,
            name_alg,
            attributes,
        )
        + auth_policy
        + parameters
        + unique
    )
    return _pack_in_tpm2b(tpmt_pub)


def _get_hasher_from_name_alg(nameAlg: int):
    if nameAlg == TPM_ALG_SHA1:
        return hashlib.sha1()
    if nameAlg == TPM_ALG_SHA256:
        return hashlib.sha256()
    if nameAlg == TPM_ALG_SHA384:
        return hashlib.sha384()
    if nameAlg == TPM_ALG_SHA512:
        return hashlib.sha512()

    raise ValueError(f"Unsupported nameAlg {nameAlg} used")


def get_tpm2b_public_object_attributes(public: bytes) -> int:
    # Ignore length, type, namealg and get attributes
    (
        _,
        _,
        _,
        attrs,
    ) = struct.unpack(">HHHI", public[0:10])
    return attrs


def get_tpm2b_public_name(public: bytes) -> str:
    """
    Return the TPM name of an object provided as TPM2B_PUBLIC.

    The name is equal to: nameAlg || H(T_Public)
    where
        H is the hash function identified by nameAlg
        T_Public is the TPMT_PUBLIC part of the object
    """
    # We get a TPM2B_PUBLIC, but don't care about the buffer portion.
    # Thus we drop the first two bytes (the uint16 size)
    (tpmt_public, rest) = _extract_tpm2b(public)
    if len(rest) != 0:
        raise ValueError("Invalid tpm2b_public")
    # The first two bytes are type, those are not critical for computing the
    # of Name other than that they are used in the computation.
    # Next two are nameAlg, which are critical.
    (nameAlg,) = struct.unpack(">H", tpmt_public[2:4])
    # Compute the H(TPMT_Public) portion
    hasher = _get_hasher_from_name_alg(int(nameAlg))
    hasher.update(tpmt_public)
    # The name is nameAlg || H(TP_PUBLIC)
    name = tpmt_public[2:4] + hasher.digest()
    # We return it as hex-encoded, since that's the way we pass it onwards
    return name.hex()


def object_attributes_description(oas: int) -> str:
    attrs = []
    if (oas & OA_FIXEDTPM) != 0:
        attrs.append("fixed-tpm")
    if (oas & OA_STCLEAR) != 0:
        attrs.append("st-clear")
    if (oas & OA_FIXEDPARENT) != 0:
        attrs.append("fixed-parent")
    if (oas & OA_SENSITIVEDATAORIGIN) != 0:
        attrs.append("sensitive-data-origin")
    if (oas & OA_USERWITHAUTH) != 0:
        attrs.append("user-with-auth")
    if (oas & OA_ADMINWITHPOLICY) != 0:
        attrs.append("admin-with-policy")
    if (oas & OA_NODA) != 0:
        attrs.append("no-da")
    if (oas & OA_ENCRYPTEDDUPLICATION) != 0:
        attrs.append("encrypted-duplication")
    if (oas & OA_RESTRICTED) != 0:
        attrs.append("restricted")
    if (oas & OA_DECRYPT) != 0:
        attrs.append("decrypt")
    if (oas & OA_SIGN_ENCRYPT) != 0:
        attrs.append("sign-encrypt")

    return " | ".join(attrs)


def ek_low_tpm2b_public_from_pubkey(pubkey: pubkey_type) -> bytes:
    return tpm2b_public_from_pubkey(
        pubkey,
        EK_LOW_NAMEALG,
        EK_LOW_ATTRIBUTES,
        EK_LOW_AUTH_POLICY,
        EK_LOW_NON_ASYM_ALG_PARMS,
    )
